"""This module is a custom module designed for general use in the Omnis project.
"""
import csv
import os

import cv2
import keras
import numpy as np


class PredictGenerator(keras.utils.Sequence):
    def __init__(self, image_filenames=None, image_classes=None,
                 batch_size=None, image_shape=None, shuffle=False):
        self.image_filenames = image_filenames
        self.image_classes = image_classes
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle
        self.predict_order_filenames = []

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.image_filenames[
            idx * self.batch_size
            :(idx + 1) * self.batch_size
        ]

        x = []
        for filename in batch_x:
            _, only_file_name = os.path.split(filename)
            self.predict_order_filenames.append(only_file_name)
            img = cv2.imread(filename)
            test_data = np.expand_dims(img, axis=0)
            resized_img = cv2.resize(test_data[0], self.image_shape)
            x.append(resized_img)

        data_array = np.asarray(x)
        data_array = data_array.astype('float32')
        data_array /= 255
        return data_array

    def flow_one_directory(self, directory_path,
                           shuffle=False, batch_size=1, image_shape=None):
        file_list = os.listdir(directory_path)
        self.image_filenames = []
        for file_name in file_list:
            self.image_filenames.append(directory_path + '/' + file_name)
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle

        return self


class TrainGenerator(keras.utils.Sequence):
    def __init__(self, image_filenames=None, image_shape=None, shuffle=False):
        self.image_filenames = image_filenames
        self.image_classes = []
        self.class_indices = []
        self.image_shape = image_shape
        self.shuffle = shuffle

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.image_filenames[
            idx * self.batch_size
            :(idx + 1) * self.batch_size
        ]
        class_x = self.image_classes[
            idx * self.batch_size
            :(idx + 1) * self.batch_size
        ]

        cnt = 0
        x = []
        index_list = []
        for filename in batch_x:
            try:
                img = cv2.imread(filename)
                resized_img = cv2.resize(img, self.image_shape)
                x.append(resized_img)
                index_list.append(cnt)
                cnt += 1
            except BaseException:
                pass

        batch_y = np.zeros((cnt, len(self.class_indices)), dtype='float')

        i = 0
        for index in index_list:
            find_index = self.class_indices.index(class_x[index])
            batch_y[i, find_index] = 1
            i += 1

        x_array = np.array(x)
        return x_array, batch_y

    def flow_from_csv(self, csv_file_path, image_directory_path,
                      shuffle=False, target_size=None, batch_size=1):
        f = open(csv_file_path, 'r', encoding='utf-8')
        csv_reader = csv.reader(f)
        self.image_classes = list()
        self.image_filenames = list()

        for line in csv_reader:
            atoms = line
            file_full_path = os.path.join(
                image_directory_path, atoms[0] + '.jpg')
            if os.path.isfile(file_full_path):
                self.image_classes.append(atoms[2])
                self.image_filenames.append(file_full_path)
        f.close()
        self.class_indices = list(set(self.image_classes))
        self.shuffle = shuffle
        self.image_shape = target_size
        self.batch_size = batch_size
        self.n = len(self.image_filenames)
        return self



import json
from collections import OrderedDict

from keras.callbacks import Callback


class ImageClassificationCallback(Callback):
    def __init__(self, total_step, total_epoch):
        Callback.__init__(self)
        self.step = 1
        self.epoch = 1
        self.total_step = total_step
        self.total_epoch = total_epoch

    def on_batch_end(self, batch, logs=None):
        if logs is None: logs = {}
        log_json = OrderedDict()
        log_json['logType'] = "batchend"
        log_json['totalStep'] = self.total_step
        log_json['nowStep'] = self.step
        log_json['totalEpoch'] = self.total_epoch
        log_json['nowEpoch'] = self.epoch
        log_json['loss'] = str(logs.get('loss'))
        log_json['acc'] = str(logs.get('acc'))
        print(json.dumps(log_json, ensure_ascii=False))
        self.step += 1

    def on_epoch_end(self, epoch, logs=None):
        if logs is None: logs = {}
        self.step = 1
        log_json = OrderedDict()
        log_json['logType'] = "epochend"
        log_json['totalEpoch'] = self.total_epoch
        log_json['nowEpoch'] = self.epoch
        log_json['loss'] = str(logs.get('loss'))
        log_json['acc'] = str(logs.get('acc'))
        print(json.dumps(log_json, ensure_ascii=False))
        self.epoch += 1



from __future__ import division, print_function

from collections import OrderedDict
import csv
import json
import os

import cv2
import keras
from keras.preprocessing.image import ImageDataGenerator
from keras.utils import multi_gpu_model
from math import ceil
import numpy as np

from ...application import Application
from ....lib.config_lib import DeepblockLogConfig, ImageClassificationConfig
from ....lib.custom_callback import ImageClassificationCallback
from ....lib.general_lib import (
    get_data_path_type, reverse_dict,
)
from ....lib.image_lib import reshape_data
from ....lib.model_lib import image_classification_model
from ....lib.saving_lib import (
    load_models_with_class_indices, save_models_with_class_indices,
)


class ImageClassification(Application):
    def __init__(self, model_path=None, model_type='densenet121'):
        super().__init__()

        self.model = None
        self.model_type = model_type

        if model_path is not None:
            self.load(model_path)


    def train(self,
              data_path=None,
              optimizer='nadam',
              metrics='accuracy',
              epochs=1,
              batch_size=32,
              verbose=0,
              shuffle=False):
        if not isinstance(metrics, list):
            metrics = [metrics]
        get_image_from = get_data_path_type(data_path)

        train_generator = self.make_train_generator(
            get_image_from=get_image_from,
            data_path=data_path,
            batch_size=batch_size)
        class_indices = self.change_dict_value_int_from_string(
            train_generator.class_indices)
        num_dataset = train_generator.n
        steps_per_epoch = ceil(num_dataset / batch_size)

        if self.model is None:
            self.model = self.create_and_compile_model(
                class_indices=class_indices,
                model_type=self.model_type,
                optimizer=optimizer,
                metrics=metrics)
        if self.deepblock_log:
            deepblock_log_json = OrderedDict()
            deepblock_log_json[
                DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.TRAIN_START
            print(json.dumps(deepblock_log_json, ensure_ascii=False))
        callbacks = []
        if self.deepblock_log:
            callbacks = [ImageClassificationCallback(
                total_step=steps_per_epoch, total_epoch=epochs)]

        self.model.fit_generator(
            train_generator,
            steps_per_epoch=steps_per_epoch,
            epochs=epochs,
            verbose=verbose,
            callbacks=callbacks,
            shuffle=shuffle)

        if self.deepblock_log:
            deepblock_log_json = OrderedDict()
            deepblock_log_json[DeepblockLogConfig.LOG_TYPE] = \
                DeepblockLogConfig.TRAIN_END
            print(json.dumps(deepblock_log_json, ensure_ascii=False))

    def predict(self, data_path, batch_size=16, verbose=0):
        data_path_type = get_data_path_type(data_path)

        pair_dict = OrderedDict()

        if self.deepblock_log:
            deepblock_log_json = OrderedDict()
            deepblock_log_json[
                DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.PREDICT_START
            print(json.dumps(deepblock_log_json, ensure_ascii=False))

        if data_path_type == 'directory':

            datagen = OneDirectoryImageGenerator()
            predict_data_generator = datagen.flow_one_directory(
                directory_path=data_path,
                shuffle=False,
                batch_size=batch_size,
                image_shape=self.model.input_shape[1:3]
            )

            num_dataset = len(predict_data_generator.image_filenames)
            probs = self.model.predict_generator(
                predict_data_generator, steps=ceil(num_dataset / batch_size))
            predicted_order_filenames = \
                predict_data_generator.predict_order_filenames
            predicted_class_indices = probs.argmax(axis=-1)

            reversed_dict = reverse_dict(self.model.class_indices)
            for i in range(len(predicted_class_indices)):
                pair_dict[predicted_order_filenames[i]] = reversed_dict[
                    str(probs.argmax(axis=-1)[i])]

            if self.deepblock_log:
                log_dict = OrderedDict()
                log_dict[
                    DeepblockLogConfig.LOG_TYPE] = \
                    DeepblockLogConfig.PREDICT_END
                log_dict[DeepblockLogConfig.RESULT] = pair_dict
                return json.dumps(log_dict, ensure_ascii=False)
            else:
                return json.dumps(pair_dict, ensure_ascii=False)
        elif data_path_type == 'image':
            _, only_file_name = os.path.split(data_path)
            img_to_predict = cv2.imread(data_path)
            img_array = np.expand_dims(img_to_predict, axis=0)
            reshaped_array = reshape_data(data_array=img_array,
                                          input_shape=self.model.input_shape)
            probs = self.model.predict(
                reshaped_array,
                batch_size=batch_size,
                verbose=verbose)

            reversed_dict = reverse_dict(self.model.class_indices)
            pair_dict[only_file_name] = reversed_dict[
                str(probs.argmax(axis=-1)[0])]

            if self.deepblock_log:
                log_dict = OrderedDict()
                log_dict[
                    DeepblockLogConfig.LOG_TYPE] = \
                    DeepblockLogConfig.PREDICT_END
                log_dict[DeepblockLogConfig.RESULT] = pair_dict
                return json.dumps(log_dict, ensure_ascii=False)
            else:
                return json.dumps(pair_dict, ensure_ascii=False)
        else:
            assert False
            return None

    def load(self, model_path=None):
        try:
            self.model = load_models_with_class_indices(model_path)
            if self.gpu_num > 1:
                self.model = multi_gpu_model(self.model, gpus=self.gpu_num)
        except Exception as e:
            raise e

    def save(self, model_path):
        try:
            if isinstance(self.model, type(None)):
                raise TypeError('You should create a model before saving it')

            if self.deepblock_log:
                log_json = OrderedDict()
                log_json[
                    DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.SAVE_START
                log_json[DeepblockLogConfig.SUCCESS] = "true"
                print(json.dumps(log_json, ensure_ascii=False))

            save_models_with_class_indices(self.model, model_path)

            if self.deepblock_log:
                log_json = OrderedDict()
                log_json[
                    DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.SAVE_END
                log_json[DeepblockLogConfig.SUCCESS] = "true"
                print(json.dumps(log_json, ensure_ascii=False))
        except Exception as e:
            raise e

    def create_and_compile_model(self, class_indices, model_type, optimizer,
                                 metrics):
        model = image_classification_model(num_classes=len(class_indices),
                                           gpu_num=self.gpu_num,
                                           model_type=model_type)
        model.compile(optimizer=optimizer,
                      loss=keras.losses.categorical_crossentropy,
                      metrics=metrics)
        model.class_indices = class_indices
        return model

    def make_train_generator(self, get_image_from, data_path, batch_size):
        if get_image_from == 'directory':
            train_datagen = ImageDataGenerator(rescale=1. / 255.)
            train_generator = train_datagen.flow_from_directory(
                data_path,
                target_size=ImageClassificationConfig.INPUT_SHAPE[0:2],
                class_mode='categorical',
                batch_size=batch_size)
        elif get_image_from == 'csv':
            train_datagen = CsvGenerator()
            train_generator = train_datagen.flow_from_csv(
                csv_file_path=data_path,
                image_directory_path="",
                target_size=ImageClassificationConfig.INPUT_SHAPE[0:2],
                batch_size=batch_size)

        return train_generator

    def evaluate(self, data_path=None, batch_size=1):
        try:
            if isinstance(data_path, type(None)):
                raise TypeError('You should give a proper data path')

            evaluate_datagen = ImageDataGenerator(rescale=1. / 255)
            validation_generator = evaluate_datagen.flow_from_directory(
                data_path,
                target_size=self.model.input_shape[1:3],
                batch_size=batch_size
            )

            scores = self.model.evaluate_generator(
                validation_generator,
                ceil(len(validation_generator) / batch_size))
            print("accuracy : ", scores[1] * 100)
        except Exception as e:
            raise e

    def change_dict_value_int_from_string(self, class_indices):
        string_class_indices = {}
        for key, value in class_indices.items():
            string_class_indices[key] = str(value)

        return string_class_indices


class CsvGenerator(keras.utils.Sequence):
    def __init__(self, image_filenames=None, image_shape=None, shuffle=False):
        self.image_filenames = image_filenames
        self.image_classes = []
        self.class_indices = []
        self.image_shape = image_shape
        self.shuffle = shuffle
        self.batch_size = 1
        self.n = 0

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.image_filenames[
                  idx * self.batch_size
                  :(idx + 1) * self.batch_size
                  ]
        class_x = self.image_classes[
                  idx * self.batch_size
                  :(idx + 1) * self.batch_size
                  ]

        cnt = 0
        x = []
        index_list = []
        for filename in batch_x:
            try:
                img = cv2.imread(filename)
                resized_img = cv2.resize(img, self.image_shape)
                x.append(resized_img)
                index_list.append(cnt)
                cnt += 1
            except BaseException:
                pass

        batch_y = np.zeros((cnt, len(self.class_indices)), dtype='float')

        i = 0
        for index in index_list:
            find_index = self.class_indices.index(class_x[index])
            batch_y[i, find_index] = 1
            i += 1

        x_array = np.array(x)
        return x_array, batch_y

    def flow_from_csv(self, csv_file_path, image_directory_path,
                      shuffle=False, target_size=None, batch_size=1):
        f = open(csv_file_path, 'r', encoding='utf-8')
        csv_reader = csv.reader(f)
        self.image_classes = list()
        self.image_filenames = list()

        for line in csv_reader:
            atoms = line
            file_full_path = os.path.join(
                image_directory_path, atoms[0] + '.jpg')
            if os.path.isfile(file_full_path):
                self.image_classes.append(atoms[2])
                self.image_filenames.append(file_full_path)
        f.close()
        self.class_indices = list(set(self.image_classes))
        self.shuffle = shuffle
        self.image_shape = target_size
        self.batch_size = batch_size
        self.n = len(self.image_filenames)
        return self


class OneDirectoryImageGenerator(keras.utils.Sequence):
    def __init__(self, image_filenames=None, image_classes=None,
                 batch_size=None, image_shape=None, shuffle=False):
        self.image_filenames = image_filenames
        self.image_classes = image_classes
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle
        self.predict_order_filenames = []

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        batch_x = self.image_filenames[
                  idx * self.batch_size
                  :(idx + 1) * self.batch_size
                  ]

        x = []
        for filename in batch_x:
            _, only_file_name = os.path.split(filename)
            self.predict_order_filenames.append(only_file_name)
            img = cv2.imread(filename)
            test_data = np.expand_dims(img, axis=0)
            resized_img = cv2.resize(test_data[0], self.image_shape)
            x.append(resized_img)

        data_array = np.asarray(x)
        data_array = data_array.astype('float32')
        data_array /= 255
        return data_array

    def flow_one_directory(self, directory_path,
                           shuffle=False, batch_size=1, image_shape=None):
        file_list = os.listdir(directory_path)
        self.image_filenames = [
            f'{directory_path}/{file_name}' for file_name in file_list
        ]
        self.batch_size = batch_size
        self.image_shape = image_shape
        self.shuffle = shuffle

        return self



from keras.models import Model
from keras.layers import (
    Input, Conv2D, MaxPooling2D, Dropout, UpSampling2D,
    concatenate,
)


def unet(input_size=(512, 512, 1)):
    inputs = Input(input_size)
    conv1 = Conv2D(64, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(inputs)
    conv1 = Conv2D(64, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    conv2 = Conv2D(128, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(pool1)
    conv2 = Conv2D(128, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    conv3 = Conv2D(256, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(pool2)
    conv3 = Conv2D(256, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    conv4 = Conv2D(512, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(pool3)
    conv4 = Conv2D(512, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv4)
    drop4 = Dropout(0.5)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(drop4)

    conv5 = Conv2D(1024, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(pool4)
    conv5 = Conv2D(1024, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv5)
    drop5 = Dropout(0.5)(conv5)

    up6 = Conv2D(512, 2, activation='relu', padding='same',
                 kernel_initializer='he_normal')(
                     UpSampling2D(size=(2, 2))(drop5))
    merge6 = concatenate([drop4, up6], axis=3)
    conv6 = Conv2D(512, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(merge6)
    conv6 = Conv2D(512, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv6)

    up7 = Conv2D(256, 2, activation='relu', padding='same',
                 kernel_initializer='he_normal')(
                     UpSampling2D(size=(2, 2))(conv6))
    merge7 = concatenate([conv3, up7], axis=3)
    conv7 = Conv2D(256, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(merge7)
    conv7 = Conv2D(256, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv7)

    up8 = Conv2D(128, 2, activation='relu', padding='same',
                 kernel_initializer='he_normal')(
                     UpSampling2D(size=(2, 2))(conv7))
    merge8 = concatenate([conv2, up8], axis=3)
    conv8 = Conv2D(128, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(merge8)
    conv8 = Conv2D(128, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv8)

    up9 = Conv2D(64, 2, activation='relu', padding='same',
                 kernel_initializer='he_normal')(
                     UpSampling2D(size=(2, 2))(conv8))
    merge9 = concatenate([conv1, up9], axis=3)
    conv9 = Conv2D(64, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(merge9)
    conv9 = Conv2D(64, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv9)
    conv9 = Conv2D(3, 3, activation='relu', padding='same',
                   kernel_initializer='he_normal')(conv9)
    conv10 = Conv2D(1, 1, activation='sigmoid')(conv9)

    model = Model(input=inputs, output=conv10)

    return model



"""
Mask R-CNN
Display and Visualization Functions.

Copyright (c) 2017 Matterport, Inc.
Licensed under the MIT License (see LICENSE for details)
Written by Waleed Abdulla
"""

import colorsys
import itertools
import os
import random
import sys

from matplotlib import lines, patches
from matplotlib.patches import Polygon
import matplotlib.pyplot as plt
import numpy as np
from skimage.measure import find_contours

from . import utils

# Root directory of the project
ROOT_DIR = os.path.abspath("../")

# Import Mask RCNN
sys.path.append(ROOT_DIR)  # To find local version of the library


############################################################
#  Visualization
############################################################

def display_images(images, titles=None, cols=4, cmap=None, norm=None,
                   interpolation=None):
    """Display the given set of images, optionally with titles.
    images: list or array of image tensors in HWC format.
    titles: optional. A list of titles to display with each image.
    cols: number of images per row
    cmap: Optional. Color map to use. For example, "Blues".
    norm: Optional. A Normalize instance to map values to colors.
    interpolation: Optional. Image interpolation to use for display.
    """
    titles = titles if titles is not None else [""] * len(images)
    rows = len(images) // cols + 1
    plt.figure(figsize=(14, 14 * rows // cols))
    i = 1
    for image, title in zip(images, titles):
        plt.subplot(rows, cols, i)
        plt.title(title, fontsize=9)
        plt.axis('off')
        plt.imshow(image.astype(np.uint8), cmap=cmap,
                   norm=norm, interpolation=interpolation)
        i += 1
    plt.show()


def random_colors(N, bright=True):
    """
    Generate random colors.
    To get visually distinct colors, generate them in HSV space then
    convert to RGB.
    """
    brightness = 1.0 if bright else 0.7
    hsv = [(i / N, 1, brightness) for i in range(N)]
    colors = list(map(lambda c: colorsys.hsv_to_rgb(*c), hsv))
    random.shuffle(colors)
    return colors


def apply_mask(image, mask, color, alpha=0.5):
    """Apply the given mask to the image.
    """
    for c in range(3):
        image[:, :, c] = np.where(mask == 1,
                                  image[:, :, c] *
                                  (1 - alpha) + alpha * color[c] * 255,
                                  image[:, :, c])
    return image


def display_instances(image, boxes, masks, class_ids, class_names,
                      scores=None, title="",
                      figsize=(16, 16), ax=None,
                      show_mask=True, show_bbox=True,
                      colors=None, captions=None, **kwargs):
    """
    boxes: [num_instance, (y1, x1, y2, x2, class_id)] in image coordinates.
    masks: [height, width, num_instances]
    class_ids: [num_instances]
    class_names: list of class names of the dataset
    scores: (optional) confidence scores for each box
    title: (optional) Figure title
    show_mask, show_bbox: To show masks and bounding boxes or not
    figsize: (optional) the size of the image
    colors: (optional) An array or colors to use with each object
    captions: (optional) A list of strings to use as captions for each object
    """
    # Number of instances
    N = boxes.shape[0]
    if not N:
        print("\n*** No instances to display *** \n")
    else:
        assert boxes.shape[0] == masks.shape[-1] == class_ids.shape[0]

    # If no axis is passed, create one and automatically call show()
    auto_show = False
    if not ax:
        _, ax = plt.subplots(1, figsize=figsize)
        auto_show = True

    # Generate random colors
    colors = colors or random_colors(N)

    # Show area outside image boundaries.
    height, width = image.shape[:2]
    ax.set_ylim(height + 10, -10)
    ax.set_xlim(-10, width + 10)
    ax.axis('off')
    ax.set_title(title)

    masked_image = image.astype(np.uint32).copy()
    contours_list = []
    for i in range(N):
        color = colors[i]

        # Bounding box
        if not np.any(boxes[i]):
            # Skip this instance. Has no bbox. Likely lost in image cropping.
            continue
        y1, x1, y2, x2 = boxes[i]
        if show_bbox:
            p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2,
                                  alpha=0.7, linestyle="dashed",
                                  edgecolor=color, facecolor='none')
            ax.add_patch(p)

        # Label
        if not captions:
            class_id = class_ids[i]
            score = scores[i] if scores is not None else None
            label = class_names[class_id]
            x = random.randint(x1, (x1 + x2) // 2)
            caption = "{} {:.3f}".format(label, score) if score else label
        else:
            caption = captions[i]
        ax.text(x1, y1 + 8, caption,
                color='w', size=11, backgroundcolor="none")

        # Mask
        mask = masks[:, :, i]
        if show_mask:
            masked_image = apply_mask(masked_image, mask, color)

        # Mask Polygon
        # Pad to ensure proper polygons for masks that touch image edges.
        padded_mask = np.zeros(
            (mask.shape[0] + 2, mask.shape[1] + 2), dtype=np.uint8)
        padded_mask[1:-1, 1:-1] = mask
        contours = find_contours(padded_mask, 0.5)
        for verts in contours:
            # Subtract the padding and flip (y, x) to (x, y)
            verts = np.fliplr(verts) - 1
            contours_list.append(verts)
            p = Polygon(verts, facecolor="none", edgecolor=color)
            ax.add_patch(p)
    ax.imshow(masked_image.astype(np.uint8))
    # if auto_show or show_masked_image:
    #     plt.show()
    return plt, contours_list


def display_differences(image,
                        gt_box, gt_class_id, gt_mask,
                        pred_box, pred_class_id, pred_score, pred_mask,
                        class_names, title="", ax=None,
                        show_mask=True, show_box=True,
                        iou_threshold=0.5, score_threshold=0.5):
    """Display ground truth and prediction instances on the same image."""
    # Match predictions to ground truth
    gt_match, pred_match, overlaps = utils.compute_matches(
        gt_box, gt_class_id, gt_mask,
        pred_box, pred_class_id, pred_score, pred_mask,
        iou_threshold=iou_threshold, score_threshold=score_threshold)
    # Ground truth = green. Predictions = red
    colors = [(0, 1, 0, .8)] * len(gt_match) \
             + [(1, 0, 0, 1)] * len(pred_match)
    # Concatenate GT and predictions
    class_ids = np.concatenate([gt_class_id, pred_class_id])
    scores = np.concatenate([np.zeros([len(gt_match)]), pred_score])
    boxes = np.concatenate([gt_box, pred_box])
    masks = np.concatenate([gt_mask, pred_mask], axis=-1)
    # Captions per instance show score/IoU
    captions = ["" for m in gt_match] + ["{:.2f} / {:.2f}".format(
        pred_score[i],
        (overlaps[i, int(pred_match[i])]
         if pred_match[i] > -1 else overlaps[i].max()))
                                         for i in range(len(pred_match))]
    # Set title if not provided
    title = title or "Ground Truth and Detections\n GT=green, pred=red, captions: score/IoU"
    # Display
    display_instances(
        image,
        boxes, masks, class_ids,
        class_names, scores, ax=ax,
        show_bbox=show_box, show_mask=show_mask,
        colors=colors, captions=captions,
        title=title)


def draw_rois(image, rois, refined_rois, mask,
              class_ids, class_names, limit=10):
    """
    anchors: [n, (y1, x1, y2, x2)] list of anchors in image coordinates.
    proposals: [n, 4] the same anchors but refined to fit objects better.
    """
    masked_image = image.copy()

    # Pick random anchors in case there are too many.
    ids = np.arange(rois.shape[0], dtype=np.int32)
    ids = np.random.choice(
        ids, limit, replace=False) if ids.shape[0] > limit else ids

    fig, ax = plt.subplots(1, figsize=(12, 12))
    if rois.shape[0] > limit:
        plt.title("Showing {} random ROIs out of {}".format(
            len(ids), rois.shape[0]))
    else:
        plt.title("{} ROIs".format(len(ids)))

    # Show area outside image boundaries.
    ax.set_ylim(image.shape[0] + 20, -20)
    ax.set_xlim(-50, image.shape[1] + 20)
    ax.axis('off')

    for i, id in enumerate(ids):
        color = np.random.rand(3)
        class_id = class_ids[id]
        # ROI
        y1, x1, y2, x2 = rois[id]
        p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2,
                              edgecolor=color if class_id else "gray",
                              facecolor='none', linestyle="dashed")
        ax.add_patch(p)
        # Refined ROI
        if class_id:
            ry1, rx1, ry2, rx2 = refined_rois[id]
            p = patches.Rectangle(
                (rx1,
                 ry1),
                rx2 - rx1,
                ry2 - ry1,
                linewidth=2,
                edgecolor=color,
                facecolor='none')
            ax.add_patch(p)
            # Connect the top-left corners of the anchor and proposal for easy
            # visualization
            ax.add_line(lines.Line2D([x1, rx1], [y1, ry1], color=color))

            # Label
            label = class_names[class_id]
            ax.text(rx1, ry1 + 8, "{}".format(label),
                    color='w', size=11, backgroundcolor="none")

            # Mask
            m = utils.unmold_mask(mask[id], rois[id]
                                  [:4].astype(np.int32), image.shape)
            masked_image = apply_mask(masked_image, m, color)

    ax.imshow(masked_image)

    # Print stats
    print("Positive ROIs: ", class_ids[class_ids > 0].shape[0])
    print("Negative ROIs: ", class_ids[class_ids == 0].shape[0])
    print("Positive Ratio: {:.2f}".format(
        class_ids[class_ids > 0].shape[0] / class_ids.shape[0]))


# TODO: Replace with matplotlib equivalent?
def draw_box(image, box, color):
    """Draw 3-pixel width bounding boxes on the given image array.
    color: list of 3 int values for RGB.
    """
    y1, x1, y2, x2 = box
    image[y1:y1 + 2, x1:x2] = color
    image[y2:y2 + 2, x1:x2] = color
    image[y1:y2, x1:x1 + 2] = color
    image[y1:y2, x2:x2 + 2] = color
    return image


def display_top_masks(image, mask, class_ids, class_names, limit=4):
    """Display the given image and the top few class masks."""
    to_display = []
    titles = []
    to_display.append(image)
    titles.append("H x W={}x{}".format(image.shape[0], image.shape[1]))
    # Pick top prominent classes in this image
    unique_class_ids = np.unique(class_ids)
    mask_area = [np.sum(mask[:, :, np.where(class_ids == i)[0]])
                 for i in unique_class_ids]
    top_ids = [v[0] for v in sorted(zip(unique_class_ids, mask_area),
                                    key=lambda r: r[1], reverse=True) if
               v[1] > 0]
    # Generate images and titles
    for i in range(limit):
        class_id = top_ids[i] if i < len(top_ids) else -1
        # Pull masks of instances belonging to the same class.
        m = mask[:, :, np.where(class_ids == class_id)[0]]
        m = np.sum(m * np.arange(1, m.shape[-1] + 1), -1)
        to_display.append(m)
        titles.append(class_names[class_id] if class_id != -1 else "-")
    display_images(to_display, titles=titles, cols=limit + 1, cmap="Blues_r")


def plot_precision_recall(AP, precisions, recalls):
    """Draw the precision-recall curve.

    AP: Average precision at IoU >= 0.5
    precisions: list of precision values
    recalls: list of recall values
    """
    # Plot the Precision-Recall curve
    _, ax = plt.subplots(1)
    ax.set_title("Precision-Recall Curve. AP@50 = {:.3f}".format(AP))
    ax.set_ylim(0, 1.1)
    ax.set_xlim(0, 1.1)
    _ = ax.plot(recalls, precisions)


def plot_overlaps(gt_class_ids, pred_class_ids, pred_scores,
                  overlaps, class_names, threshold=0.5):
    """Draw a grid showing how ground truth objects are classified.
    gt_class_ids: [N] int. Ground truth class IDs
    pred_class_id: [N] int. Predicted class IDs
    pred_scores: [N] float. The probability scores of predicted classes
    overlaps: [pred_boxes, gt_boxes] IoU overlaps of predictions and GT boxes.
    class_names: list of all class names in the dataset
    threshold: Float. The prediction probability required to predict a class
    """
    gt_class_ids = gt_class_ids[gt_class_ids != 0]
    pred_class_ids = pred_class_ids[pred_class_ids != 0]

    plt.figure(figsize=(12, 10))
    plt.imshow(overlaps, interpolation='nearest', cmap=plt.cm.Blues)
    plt.yticks(np.arange(len(pred_class_ids)),
               ["{} ({:.2f})".format(class_names[int(id)], pred_scores[i])
                for i, id in enumerate(pred_class_ids)])
    plt.xticks(np.arange(len(gt_class_ids)),
               [class_names[int(id)] for id in gt_class_ids], rotation=90)

    thresh = overlaps.max() / 2.
    for i, j in itertools.product(range(overlaps.shape[0]),
                                  range(overlaps.shape[1])):
        text = ""
        if overlaps[i, j] > threshold:
            text = "match" if gt_class_ids[j] == pred_class_ids[i] else "wrong"
        color = ("white" if overlaps[i, j] > thresh
                 else "black" if overlaps[i, j] > 0
                 else "grey")
        plt.text(j, i, "{:.3f}\n{}".format(overlaps[i, j], text),
                 horizontalalignment="center", verticalalignment="center",
                 fontsize=9, color=color)

    plt.tight_layout()
    plt.xlabel("Ground Truth")
    plt.ylabel("Predictions")


def draw_boxes(image, boxes=None, refined_boxes=None,
               masks=None, captions=None, visibilities=None,
               title="", ax=None):
    """Draw bounding boxes and segmentation masks with different
    customizations.

    boxes: [N, (y1, x1, y2, x2, class_id)] in image coordinates.
    refined_boxes: Like boxes, but draw with solid lines to show
        that they're the result of refining 'boxes'.
    masks: [N, height, width]
    captions: List of N titles to display on each box
    visibilities: (optional) List of values of 0, 1, or 2. Determine how
        prominent each bounding box should be.
    title: An optional title to show over the image
    ax: (optional) Matplotlib axis to draw on.
    """
    # Number of boxes
    assert boxes is not None or refined_boxes is not None
    N = boxes.shape[0] if boxes is not None else refined_boxes.shape[0]

    # Matplotlib Axis
    if not ax:
        _, ax = plt.subplots(1, figsize=(12, 12))

    # Generate random colors
    colors = random_colors(N)

    # Show area outside image boundaries.
    margin = image.shape[0] // 10
    ax.set_ylim(image.shape[0] + margin, -margin)
    ax.set_xlim(-margin, image.shape[1] + margin)
    ax.axis('off')

    ax.set_title(title)

    masked_image = image.astype(np.uint32).copy()
    for i in range(N):
        # Box visibility
        visibility = visibilities[i] if visibilities is not None else 1
        if visibility == 0:
            color = "gray"
            style = "dotted"
            alpha = 0.5
        elif visibility == 1:
            color = colors[i]
            style = "dotted"
            alpha = 1
        elif visibility == 2:
            color = colors[i]
            style = "solid"
            alpha = 1

        # Boxes
        if boxes is not None:
            if not np.any(boxes[i]):
                # Skip this instance. Has no bbox. Likely lost in cropping.
                continue
            y1, x1, y2, x2 = boxes[i]
            p = patches.Rectangle((x1, y1), x2 - x1, y2 - y1, linewidth=2,
                                  alpha=alpha, linestyle=style,
                                  edgecolor=color, facecolor='none')
            ax.add_patch(p)

        # Refined boxes
        if refined_boxes is not None and visibility > 0:
            ry1, rx1, ry2, rx2 = refined_boxes[i].astype(np.int32)
            p = patches.Rectangle(
                (rx1,
                 ry1),
                rx2 - rx1,
                ry2 - ry1,
                linewidth=2,
                edgecolor=color,
                facecolor='none')
            ax.add_patch(p)
            # Connect the top-left corners of the anchor and proposal
            if boxes is not None:
                ax.add_line(lines.Line2D([x1, rx1], [y1, ry1], color=color))

        # Captions
        if captions is not None:
            caption = captions[i]
            # If there are refined boxes, display captions on them
            if refined_boxes is not None:
                y1, x1, y2, x2 = ry1, rx1, ry2, rx2
            x = random.randint(x1, (x1 + x2) // 2)
            ax.text(x1, y1, caption, size=11, verticalalignment='top',
                    color='w', backgroundcolor="none",
                    bbox={
                        'facecolor': color, 'alpha': 0.5,
                        'pad': 2, 'edgecolor': 'none'
                    })

        # Masks
        if masks is not None:
            mask = masks[:, :, i]
            masked_image = apply_mask(masked_image, mask, color)
            # Mask Polygon
            # Pad to ensure proper polygons for masks that touch image edges.
            padded_mask = np.zeros(
                (mask.shape[0] + 2, mask.shape[1] + 2), dtype=np.uint8)
            padded_mask[1:-1, 1:-1] = mask
            contours = find_contours(padded_mask, 0.5)
            for verts in contours:
                # Subtract the padding and flip (y, x) to (x, y)
                verts = np.fliplr(verts) - 1
                p = Polygon(verts, facecolor="none", edgecolor=color)
                ax.add_patch(p)
    ax.imshow(masked_image.astype(np.uint8))


def display_table(table):
    """Display values in a table format.
    table: an iterable of rows, and each row is an iterable of values.
    """
    html = ""
    for row in table:
        row_html = ""
        for col in row:
            row_html += "<td>{:40}</td>".format(str(col))
        html += "<tr>" + row_html + "</tr>"
    html = "<table>" + html + "</table>"
    # IPython.display.display(IPython.display.HTML(html))


def display_weight_stats(model):
    """Scans all the weights in the model and returns a list of tuples
    that contain stats about each weight.
    """
    layers = model.get_trainable_layers()
    table = [["WEIGHT NAME", "SHAPE", "MIN", "MAX", "STD"]]
    for l in layers:
        weight_values = l.get_weights()  # list of Numpy arrays
        weight_tensors = l.weights  # list of TF tensors
        for i, w in enumerate(weight_values):
            weight_name = weight_tensors[i].name
            # Detect problematic layers. Exclude biases of conv layers.
            alert = ""
            if w.min() == w.max() and not (
                    l.__class__.__name__ == "Conv2D" and i == 1):
                alert += "<span style='color:red'>*** dead?</span>"
            if np.abs(w.min()) > 1000 or np.abs(w.max()) > 1000:
                alert += "<span style='color:red'>*** Overflow?</span>"
            # Add row
            table.append([
                weight_name + alert,
                str(w.shape),
                "{:+9.4f}".format(w.min()),
                "{:+10.4f}".format(w.max()),
                "{:+9.4f}".format(w.std()),
            ])
    display_table(table)



"""Miscellaneous utility functions."""

from functools import reduce

import numpy as np
from PIL import Image
from matplotlib.colors import hsv_to_rgb, rgb_to_hsv


def compose(*funcs):
    """Compose arbitrarily many functions, evaluated left to right.

    Reference: https://mathieularose.com/function-composition-in-python/
    """
    # return lambda x: reduce(lambda v, f: f(v), funcs, x)
    if funcs:
        return reduce(lambda f, g: lambda *a, **kw: g(f(*a, **kw)), funcs)
    else:
        raise ValueError('Composition of empty sequence not supported.')


def letterbox_image(image, size):
    '''resize image with unchanged aspect ratio using padding'''
    iw, ih = image.size
    w, h = size
    scale = min(w / iw, h / ih)
    nw = int(iw * scale)
    nh = int(ih * scale)

    image = image.resize((nw, nh), Image.BICUBIC)
    new_image = Image.new('RGB', size, (128, 128, 128))
    new_image.paste(image, ((w - nw) // 2, (h - nh) // 2))
    return new_image


def rand(a=0, b=1):
    return np.random.rand() * (b - a) + a


def get_random_data(
        annotation_line,
        input_shape,
        random=True,
        max_boxes=20,
        jitter=.3,
        hue=.1,
        sat=1.5,
        val=1.5,
        proc_img=True):
    '''random preprocessing for real-time data augmentation'''
    line = annotation_line.split()  # split by SPACE. -libn
    image = Image.open(line[0])
    iw, ih = image.size
    h, w = input_shape
    box = np.array([np.array(list(map(int, box.split(','))))
                    for box in line[1:]])

    if not random:
        # resize image
        scale = min(w / iw, h / ih)
        nw = int(iw * scale)
        nh = int(ih * scale)
        dx = (w - nw) // 2
        dy = (h - nh) // 2
        image_data = 0
        if proc_img:
            image = image.resize((nw, nh), Image.BICUBIC)
            new_image = Image.new('RGB', (w, h), (128, 128, 128))
            new_image.paste(image, (dx, dy))
            image_data = np.array(new_image) / 255.

        # correct boxes
        box_data = np.zeros((max_boxes, 5))
        if box:
            np.random.shuffle(box)
            if len(box) > max_boxes:
                box = box[:max_boxes]
            box[:, [0, 2]] = box[:, [0, 2]] * scale + dx
            box[:, [1, 3]] = box[:, [1, 3]] * scale + dy
            box_data[:len(box)] = box

        return image_data, box_data

    # resize image
    new_ar = w / h * rand(1 - jitter, 1 + jitter) / \
             rand(1 - jitter, 1 + jitter)
    scale = rand(.25, 2)
    if new_ar < 1:
        nh = int(scale * h)
        nw = int(nh * new_ar)
    else:
        nw = int(scale * w)
        nh = int(nw / new_ar)
    image = image.resize((nw, nh), Image.BICUBIC)

    # place image
    dx = int(rand(0, w - nw))
    dy = int(rand(0, h - nh))
    new_image = Image.new('RGB', (w, h), (128, 128, 128))
    new_image.paste(image, (dx, dy))
    image = new_image

    # flip image or not
    flip = rand() < .5
    if flip:
        image = image.transpose(Image.FLIP_LEFT_RIGHT)

    # distort image
    hue = rand(-hue, hue)
    sat = rand(1, sat) if rand() < .5 else 1 / rand(1, sat)
    val = rand(1, val) if rand() < .5 else 1 / rand(1, val)
    x = rgb_to_hsv(np.array(image) / 255.)
    x[..., 0] += hue
    x[..., 0][x[..., 0] > 1] -= 1
    x[..., 0][x[..., 0] < 0] += 1
    x[..., 1] *= sat
    x[..., 2] *= val
    x[x > 1] = 1
    x[x < 0] = 0
    image_data = hsv_to_rgb(x)  # numpy array, 0 to 1

    # correct boxes
    box_data = np.zeros((max_boxes, 5))
    if box:
        np.random.shuffle(box)
        box[:, [0, 2]] = box[:, [0, 2]] * nw / iw + dx
        box[:, [1, 3]] = box[:, [1, 3]] * nh / ih + dy
        if flip:
            box[:, [0, 2]] = w - box[:, [2, 0]]
        box[:, 0:2][box[:, 0:2] < 0] = 0
        box[:, 2][box[:, 2] > w] = w
        box[:, 3][box[:, 3] > h] = h
        box_w = box[:, 2] - box[:, 0]
        box_h = box[:, 3] - box[:, 1]
        box = box[np.logical_and(box_w > 1, box_h > 1)]  # discard invalid box
        if len(box) > max_boxes:
            box = box[:max_boxes]
        box_data[:len(box)] = box

    return image_data, box_data



"""YOLO_v3 Model Defined in Keras."""

from functools import wraps

from keras import backend as K
from keras.layers import (
    Add, Concatenate, Conv2D, MaxPooling2D, UpSampling2D, ZeroPadding2D,
)
from keras.layers.advanced_activations import LeakyReLU
from keras.layers.normalization import BatchNormalization
from keras.models import Model
from keras.regularizers import l2
import numpy as np
import tensorflow as tf

from .utils import compose


@wraps(Conv2D)
def DarknetConv2D(*args, **kwargs):
    """Wrapper to set Darknet parameters for Convolution2D."""
    darknet_conv_kwargs = {'kernel_regularizer': l2(5e-4)}
    darknet_conv_kwargs['padding'] = 'valid' if kwargs.get(
        'strides') == (2, 2) else 'same'
    darknet_conv_kwargs.update(kwargs)
    return Conv2D(*args, **darknet_conv_kwargs)


def DarknetConv2D_BN_Leaky(*args, **kwargs):
    """Darknet Convolution2D followed by BatchNormalization and LeakyReLU."""
    no_bias_kwargs = {'use_bias': False}
    no_bias_kwargs.update(kwargs)
    return compose(
        DarknetConv2D(*args, **no_bias_kwargs),
        BatchNormalization(),
        LeakyReLU(alpha=0.1))


def resblock_body(x, num_filters, num_blocks):
    '''A series of resblocks starting with a downsampling Convolution2D'''
    # Darknet uses left and top padding instead of 'same' mode
    x = ZeroPadding2D(((1, 0), (1, 0)))(x)
    x = DarknetConv2D_BN_Leaky(num_filters, (3, 3), strides=(2, 2))(x)
    for _ in range(num_blocks):
        y = compose(
            DarknetConv2D_BN_Leaky(num_filters // 2, (1, 1)),
            DarknetConv2D_BN_Leaky(num_filters, (3, 3)))(x)
        x = Add()([x, y])
    return x


def darknet_body(x):
    '''Darknet body having 52 Convolution2D layers'''
    x = DarknetConv2D_BN_Leaky(32, (3, 3))(x)
    x = resblock_body(x, 64, 1)
    x = resblock_body(x, 128, 2)
    x = resblock_body(x, 256, 8)
    x = resblock_body(x, 512, 8)
    x = resblock_body(x, 1024, 4)
    return x


def make_last_layers(x, num_filters, out_filters):
    '''6 Conv2D_BN_Leaky layers followed by a Conv2D_linear layer'''
    x = compose(
        DarknetConv2D_BN_Leaky(num_filters, (1, 1)),
        DarknetConv2D_BN_Leaky(num_filters * 2, (3, 3)),
        DarknetConv2D_BN_Leaky(num_filters, (1, 1)),
        DarknetConv2D_BN_Leaky(num_filters * 2, (3, 3)),
        DarknetConv2D_BN_Leaky(num_filters, (1, 1)))(x)
    y = compose(
        DarknetConv2D_BN_Leaky(num_filters * 2, (3, 3)),
        DarknetConv2D(out_filters, (1, 1)))(x)
    return x, y


def yolo_body(inputs, num_anchors, num_classes):
    """Create YOLO_V3 model CNN body in Keras."""
    darknet = Model(inputs, darknet_body(inputs))
    x, y1 = make_last_layers(
        darknet.output, 512, num_anchors * (num_classes + 5))

    x = compose(
        DarknetConv2D_BN_Leaky(256, (1, 1)),
        UpSampling2D(2))(x)
    x = Concatenate()([x, darknet.layers[152].output])
    x, y2 = make_last_layers(x, 256, num_anchors * (num_classes + 5))

    x = compose(
        DarknetConv2D_BN_Leaky(128, (1, 1)),
        UpSampling2D(2))(x)
    x = Concatenate()([x, darknet.layers[92].output])
    x, y3 = make_last_layers(x, 128, num_anchors * (num_classes + 5))

    return Model(inputs, [y1, y2, y3])


def tiny_yolo_body(inputs, num_anchors, num_classes):
    '''Create Tiny YOLO_v3 model CNN body in keras.'''
    x1 = compose(
        DarknetConv2D_BN_Leaky(16, (3, 3)),
        MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same'),
        DarknetConv2D_BN_Leaky(32, (3, 3)),
        MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same'),
        DarknetConv2D_BN_Leaky(64, (3, 3)),
        MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same'),
        DarknetConv2D_BN_Leaky(128, (3, 3)),
        MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same'),
        DarknetConv2D_BN_Leaky(256, (3, 3)))(inputs)
    x2 = compose(
        MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same'),
        DarknetConv2D_BN_Leaky(512, (3, 3)),
        MaxPooling2D(pool_size=(2, 2), strides=(1, 1), padding='same'),
        DarknetConv2D_BN_Leaky(1024, (3, 3)),
        DarknetConv2D_BN_Leaky(256, (1, 1)))(x1)
    y1 = compose(
        DarknetConv2D_BN_Leaky(512, (3, 3)),
        DarknetConv2D(num_anchors * (num_classes + 5), (1, 1)))(x2)

    x2 = compose(
        DarknetConv2D_BN_Leaky(128, (1, 1)),
        UpSampling2D(2))(x2)
    y2 = compose(
        Concatenate(),
        DarknetConv2D_BN_Leaky(256, (3, 3)),
        DarknetConv2D(num_anchors * (num_classes + 5), (1, 1)))([x2, x1])

    return Model(inputs, [y1, y2])


def yolo_head(feats, anchors, num_classes, input_shape, calc_loss=False):
    """Convert final layer features to bounding box parameters."""
    num_anchors = len(anchors)
    # Reshape to batch, height, width, num_anchors, box_params.
    anchors_tensor = K.reshape(K.constant(anchors), [1, 1, 1, num_anchors, 2])

    grid_shape = K.shape(feats)[1:3]  # height, width
    grid_y = K.tile(K.reshape(K.arange(0, stop=grid_shape[0]), [-1, 1, 1, 1]),
                    [1, grid_shape[1], 1, 1])
    grid_x = K.tile(K.reshape(K.arange(0, stop=grid_shape[1]), [1, -1, 1, 1]),
                    [grid_shape[0], 1, 1, 1])
    grid = K.concatenate([grid_x, grid_y])
    grid = K.cast(grid, K.dtype(feats))

    feats = K.reshape(feats,
                      [-1,
                       grid_shape[0],
                       grid_shape[1],
                       num_anchors,
                       num_classes + 5])

    # Adjust predictions to each spatial grid point and anchor size.
    box_xy = (K.sigmoid(feats[..., :2]) + grid) / \
             K.cast(grid_shape[::-1], K.dtype(feats))
    box_wh = K.exp(feats[..., 2:4]) * anchors_tensor / \
             K.cast(input_shape[::-1], K.dtype(feats))
    box_confidence = K.sigmoid(feats[..., 4:5])
    box_class_probs = K.sigmoid(feats[..., 5:])

    if calc_loss:
        return grid, feats, box_xy, box_wh
    return box_xy, box_wh, box_confidence, box_class_probs


def yolo_correct_boxes(box_xy, box_wh, input_shape, image_shape):
    '''Get corrected boxes'''
    box_yx = box_xy[..., ::-1]
    box_hw = box_wh[..., ::-1]
    input_shape = K.cast(input_shape, K.dtype(box_yx))
    image_shape = K.cast(image_shape, K.dtype(box_yx))
    new_shape = K.round(image_shape * K.min(input_shape / image_shape))
    offset = (input_shape - new_shape) / 2. / input_shape
    scale = input_shape / new_shape
    box_yx = (box_yx - offset) * scale
    box_hw *= scale

    box_mins = box_yx - (box_hw / 2.)
    box_maxes = box_yx + (box_hw / 2.)
    boxes = K.concatenate([
        box_mins[..., 0:1],  # y_min
        box_mins[..., 1:2],  # x_min
        box_maxes[..., 0:1],  # y_max
        box_maxes[..., 1:2]  # x_max
    ])

    # Scale boxes back to original image shape.
    boxes *= K.concatenate([image_shape, image_shape])
    return boxes


def yolo_boxes_and_scores(feats, anchors, num_classes,
                          input_shape, image_shape):
    '''Process Conv layer output'''
    box_xy, box_wh, box_confidence, box_class_probs = yolo_head(
        feats, anchors, num_classes, input_shape)
    boxes = yolo_correct_boxes(box_xy, box_wh, input_shape, image_shape)
    boxes = K.reshape(boxes, [-1, 4])
    box_scores = box_confidence * box_class_probs
    box_scores = K.reshape(box_scores, [-1, num_classes])
    return boxes, box_scores


def yolo_eval(yolo_outputs,
              anchors,
              num_classes,
              image_shape,
              max_boxes=20,
              score_threshold=.6,
              iou_threshold=.5):
    """Evaluate YOLO model on given input and return filtered boxes."""
    num_layers = len(yolo_outputs)
    anchor_mask = [[6, 7, 8], [3, 4, 5], [0, 1, 2]] if num_layers == 3 else [
        [3, 4, 5], [1, 2, 3]]  # default setting
    input_shape = K.shape(yolo_outputs[0])[1:3] * 32
    boxes = []
    box_scores = []
    for l in range(num_layers):
        boxes, box_scores = yolo_boxes_and_scores(
            yolo_outputs[l], anchors[anchor_mask[l]], num_classes, input_shape,
            image_shape)
        boxes.append(boxes)
        box_scores.append(box_scores)
    boxes = K.concatenate(boxes, axis=0)
    box_scores = K.concatenate(box_scores, axis=0)

    mask = box_scores >= score_threshold
    max_boxes_tensor = K.constant(max_boxes, dtype='int32')
    boxes_ = []
    scores_ = []
    classes_ = []
    for c in range(num_classes):
        # TODO: use keras backend instead of tf.
        class_boxes = tf.boolean_mask(boxes, mask[:, c])
        class_box_scores = tf.boolean_mask(box_scores[:, c], mask[:, c])
        nms_index = tf.image.non_max_suppression(
            class_boxes,
            class_box_scores,
            max_boxes_tensor,
            iou_threshold=iou_threshold)
        class_boxes = K.gather(class_boxes, nms_index)
        class_box_scores = K.gather(class_box_scores, nms_index)
        classes = K.ones_like(class_box_scores, 'int32') * c
        boxes_.append(class_boxes)
        scores_.append(class_box_scores)
        classes_.append(classes)
    boxes_ = K.concatenate(boxes_, axis=0)
    scores_ = K.concatenate(scores_, axis=0)
    classes_ = K.concatenate(classes_, axis=0)

    return boxes_, scores_, classes_


def preprocess_true_boxes(true_boxes, input_shape, anchors, num_classes):
    '''Preprocess true boxes to training input format

    Parameters
    ----------
    true_boxes: array, shape=(m, T, 5)
        Absolute x_min, y_min, x_max, y_max, class_id relative to input_shape.
    input_shape: array-like, hw, multiples of 32
    anchors: array, shape=(N, 2), wh
    num_classes: integer

    Returns
    -------
    y_true: list of array, shape like yolo_outputs, xywh are reletive value

    '''
    assert (true_boxes[..., 4] < num_classes).all(
    ), 'class id must be less than num_classes'
    num_layers = len(anchors) // 3  # default setting
    anchor_mask = [[6, 7, 8], [3, 4, 5], [0, 1, 2]
                   ] if num_layers == 3 else [[3, 4, 5], [1, 2, 3]]

    true_boxes = np.array(true_boxes, dtype='float32')
    input_shape = np.array(input_shape, dtype='int32')
    boxes_xy = (true_boxes[..., 0:2] + true_boxes[..., 2:4]) // 2
    boxes_wh = true_boxes[..., 2:4] - true_boxes[..., 0:2]
    true_boxes[..., 0:2] = boxes_xy / input_shape[::-1]
    true_boxes[..., 2:4] = boxes_wh / input_shape[::-1]

    m = true_boxes.shape[0]
    grid_shapes = [input_shape // {0: 32, 1: 16, 2: 8}[l]
                   for l in range(num_layers)]
    y_true = [np.zeros((m, grid_shapes[l][0], grid_shapes[l][1], len(
        anchor_mask[l]), 5 + num_classes), dtype='float32') for l in
              range(num_layers)]

    # Expand dim to apply broadcasting.
    anchors = np.expand_dims(anchors, 0)
    anchor_maxes = anchors / 2.
    anchor_mins = -anchor_maxes
    valid_mask = boxes_wh[..., 0] > 0

    for b in range(m):
        # Discard zero rows.
        wh = boxes_wh[b, valid_mask[b]]
        if not wh:
            continue
        # Expand dim to apply broadcasting.
        wh = np.expand_dims(wh, -2)
        box_maxes = wh / 2.
        box_mins = -box_maxes

        intersect_mins = np.maximum(box_mins, anchor_mins)
        intersect_maxes = np.minimum(box_maxes, anchor_maxes)
        intersect_wh = np.maximum(intersect_maxes - intersect_mins, 0.)
        intersect_area = intersect_wh[..., 0] * intersect_wh[..., 1]
        box_area = wh[..., 0] * wh[..., 1]
        anchor_area = anchors[..., 0] * anchors[..., 1]
        iou = intersect_area / (box_area + anchor_area - intersect_area)

        # Find best anchor for each true box
        best_anchor = np.argmax(iou, axis=-1)

        for t, n in enumerate(best_anchor):
            for l in range(num_layers):
                if n in anchor_mask[l]:
                    i = np.floor(true_boxes[b, t, 0] *
                                 grid_shapes[l][1]).astype('int32')
                    j = np.floor(true_boxes[b, t, 1] *
                                 grid_shapes[l][0]).astype('int32')
                    k = anchor_mask[l].index(n)
                    c = true_boxes[b, t, 4].astype('int32')
                    y_true[l][b, j, i, k, 0:4] = true_boxes[b, t, 0:4]
                    y_true[l][b, j, i, k, 4] = 1
                    y_true[l][b, j, i, k, 5 + c] = 1

    return y_true


def box_iou(b1, b2):
    '''Return iou tensor

    Parameters
    ----------
    b1: tensor, shape=(i1,...,iN, 4), xywh
    b2: tensor, shape=(j, 4), xywh

    Returns
    -------
    iou: tensor, shape=(i1,...,iN, j)

    '''

    # Expand dim to apply broadcasting.
    b1 = K.expand_dims(b1, -2)
    b1_xy = b1[..., :2]
    b1_wh = b1[..., 2:4]
    b1_wh_half = b1_wh / 2.
    b1_mins = b1_xy - b1_wh_half
    b1_maxes = b1_xy + b1_wh_half

    # Expand dim to apply broadcasting.
    b2 = K.expand_dims(b2, 0)
    b2_xy = b2[..., :2]
    b2_wh = b2[..., 2:4]
    b2_wh_half = b2_wh / 2.
    b2_mins = b2_xy - b2_wh_half
    b2_maxes = b2_xy + b2_wh_half

    intersect_mins = K.maximum(b1_mins, b2_mins)
    intersect_maxes = K.minimum(b1_maxes, b2_maxes)
    intersect_wh = K.maximum(intersect_maxes - intersect_mins, 0.)
    intersect_area = intersect_wh[..., 0] * intersect_wh[..., 1]
    b1_area = b1_wh[..., 0] * b1_wh[..., 1]
    b2_area = b2_wh[..., 0] * b2_wh[..., 1]
    iou = intersect_area / (b1_area + b2_area - intersect_area)

    return iou


def yolo_loss(args, anchors, num_classes, ignore_thresh=.5, print_loss=False):
    '''Return yolo_loss tensor

    Parameters
    ----------
    yolo_outputs: list of tensor, the output of yolo_body or tiny_yolo_body
    y_true: list of array, the output of preprocess_true_boxes
    anchors: array, shape=(N, 2), wh
    num_classes: integer
    ignore_thresh: float, the iou threshold whether to ignore object confidence loss

    Returns
    -------
    loss: tensor, shape=(1,)

    '''
    num_layers = len(anchors) // 3  # default setting
    yolo_outputs = args[:num_layers]
    y_true = args[num_layers:]
    anchor_mask = [[6, 7, 8], [3, 4, 5], [0, 1, 2]
                   ] if num_layers == 3 else [[3, 4, 5], [1, 2, 3]]
    input_shape = K.cast(
        K.shape(yolo_outputs[0])[1:3] * 32,
        K.dtype(y_true[0]))
    grid_shapes = [K.cast(K.shape(yolo_outputs[l])[1:3],
                          K.dtype(y_true[0])) for l in range(num_layers)]
    loss = 0
    m = K.shape(yolo_outputs[0])[0]  # batch size, tensor
    mf = K.cast(m, K.dtype(yolo_outputs[0]))

    for l in range(num_layers):
        object_mask = y_true[l][..., 4:5]
        true_class_probs = y_true[l][..., 5:]

        grid, raw_pred, pred_xy, pred_wh = yolo_head(
            yolo_outputs[l], anchors[anchor_mask[l]], num_classes, input_shape,
            calc_loss=True)
        pred_box = K.concatenate([pred_xy, pred_wh])

        # Darknet raw box to calculate loss.
        raw_true_xy = y_true[l][..., :2] * grid_shapes[l][::-1] - grid
        raw_true_wh = K.log(y_true[l][..., 2:4] /
                            anchors[anchor_mask[l]] * input_shape[::-1])
        raw_true_wh = K.switch(object_mask, raw_true_wh,
                               K.zeros_like(raw_true_wh))  # avoid log(0)=-inf
        box_loss_scale = 2 - y_true[l][..., 2:3] * y_true[l][..., 3:4]

        # Find ignore mask, iterate over each of batch.
        ignore_mask = tf.TensorArray(
            K.dtype(y_true[0]), size=1, dynamic_size=True)
        object_mask_bool = K.cast(object_mask, 'bool')

        def loop_body(b, ignore_mask):
            true_box = tf.boolean_mask(
                y_true[l][b, ..., 0:4], object_mask_bool[b, ..., 0])
            iou = box_iou(pred_box[b], true_box)
            best_iou = K.max(iou, axis=-1)
            ignore_mask = ignore_mask.write(
                b, K.cast(best_iou < ignore_thresh, K.dtype(true_box)))
            return b + 1, ignore_mask

        _, ignore_mask = K.control_flow_ops.while_loop(
            lambda b, *args: b < m, loop_body, [0, ignore_mask])
        ignore_mask = ignore_mask.stack()
        ignore_mask = K.expand_dims(ignore_mask, -1)

        # K.binary_crossentropy is helpful to avoid exp overflow.
        xy_loss = object_mask * box_loss_scale * \
                  K.binary_crossentropy(
                      raw_true_xy, raw_pred[..., 0:2], from_logits=True)
        wh_loss = object_mask * box_loss_scale * 0.5 * \
                  K.square(raw_true_wh - raw_pred[..., 2:4])
        crossentropy = K.binary_crossentropy(
            object_mask,
            raw_pred[..., 4:5],
            from_logits=True)
        confidence_loss = object_mask * crossentropy
        confidence_loss += object_mask * crossentropy * ignore_mask

        class_loss = object_mask * K.binary_crossentropy(
            true_class_probs, raw_pred[..., 5:], from_logits=True)

        xy_loss = K.sum(xy_loss) / mf
        wh_loss = K.sum(wh_loss) / mf
        confidence_loss = K.sum(confidence_loss) / mf
        class_loss = K.sum(class_loss) / mf
        loss += xy_loss + wh_loss + confidence_loss + class_loss
        if print_loss:
            loss = tf.Print(loss,
                            [loss,
                             xy_loss,
                             wh_loss,
                             confidence_loss,
                             class_loss,
                             K.sum(ignore_mask)],
                            message='loss: ')

        return loss



import io
import re


def read_text(text_path):
    """Reads a text file and converts it to a unicode string.

    Arguments:
        text_path {str} -- A path of text file.

    Returns:
        [unicode] -- A text file's text as a unicode.
    """
    text = io.open(text_path, encoding='utf-8').read()
    return text


def parse_sentences_and_next_characters(
        whole_text, max_sentence_size, interval_between_each_sentence):
    """Extracts sentences from whole_text according to max sentence size and
    interval between each sentence.

    Arguments:
        whole_text {str} -- A text to be parsed.
        max_sentence_size {int} -- A max length of each sentence.
        interval_between_each_sentence {int} --
            An extraction distance between each sentence.
            For example,
            if whole_text is 'hello world' and max_sentence_size is 5
            and interval_between_each_sentence is 1,
            then sentences will be
            ['hello', 'ello ', 'llo w', 'lo wo', 'o wor', ' worl', 'world'].
            or interval_between_each_sentence is 2,
            then sentences will be ['hello', 'llo w', 'o wor', 'world'].

    Returns:
        [list, list] -- A list of sentences and a list of next characters will
        be returned.
    """
    sentences = []
    next_characters = []
    for i in range(0, len(whole_text) - max_sentence_size,
                   interval_between_each_sentence):
        sentence = whole_text[i: i + max_sentence_size]
        sentences.append(sentence)
        next_character_of_sentence = whole_text[i + max_sentence_size]
        next_characters.append(next_character_of_sentence)
    return sentences, next_characters


def get_unique_characters(text):
    """Extracts unique characters from string.

    Arguments:
        text {str} -- A string from which to extract unique characters.

    Returns:
        [list] -- A list of unique characters.
    """
    unique_character_set = set(text)
    unique_characters = list(unique_character_set)
    sorted_unique_characters = sorted(unique_characters)
    return sorted_unique_characters


def get_words_from_text(text):
    word_list = re.sub(r"[^\w]", " ", text).split()  # we will lower all words
    return word_list



from keras.applications import (
    densenet, inception_v3, mobilenetv2, nasnet,
    resnet50, xception,
)
from keras.layers import (
    Activation, Add, Dropout, Input, MaxPooling2D,
    UpSampling2D,
)
from keras.layers.advanced_activations import LeakyReLU
from keras.layers.convolutional import Conv2D
from keras.layers.core import Dense, Flatten, Lambda
from keras.layers.normalization import BatchNormalization
from keras.models import Model, Sequential
from keras.utils import multi_gpu_model

from .config_lib import (
    DeblurConfig, ImageClassificationConfig,
    MultiLabelClassificationConfig,
)
from .layers_lib import ReflectionPadding2D, res_block


####################IMAGE_CLASSIFICATION_MODEL####################
def image_classification_model(num_classes, gpu_num, model_type):
    if model_type == 'densenet121':
        model = densenet.DenseNet121(weights=None, classes=num_classes,
                                     input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'inception_v3':
        model = inception_v3.InceptionV3(weights=None, classes=num_classes,
                                         input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'inception_resnet_v2':
        model = inception_v3.InceptionV3(weights=None, classes=num_classes,
                                         input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'mobilenet_v2':
        model = mobilenetv2.MobileNetV2(weights=None, classes=num_classes,
                                        input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'nasnet_large':
        model = nasnet.NASNetLarge(weights=None, classes=num_classes,
                                   input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'nasnet_mobile':
        model = nasnet.NASNetMobile(weights=None, classes=num_classes,
                                    input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'xception':
        model = xception.Xception(weights=None, classes=num_classes,
                                  input_shape=ImageClassificationConfig.INPUT_SHAPE)
    elif model_type == 'resnet50':
        model = resnet50.ResNet50(weights=None, classes=num_classes,
                                  input_shape=ImageClassificationConfig.INPUT_SHAPE)

    if gpu_num > 1:
        model = multi_gpu_model(model, gpus=gpu_num)

    return model


####################MULTI_LABEL_IMAGE_CLASSIFICATION_MODEL####################
def multi_label_classification_model(num_classes, gpu_num=1,
                                     learning_rate=0.0001):
    model = Sequential()
    model.add(Conv2D(32, (3, 3), padding='same',
                     input_shape=MultiLabelClassificationConfig.INPUT_SHAPE))
    model.add(Activation('relu'))
    model.add(Conv2D(32, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))
    model.add(Conv2D(64, (3, 3), padding='same'))
    model.add(Activation('relu'))
    model.add(Conv2D(64, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))
    model.add(Flatten())
    model.add(Dense(512))
    model.add(Activation('relu'))
    model.add(Dropout(0.5))
    model.add(Dense(num_classes, activation='sigmoid'))

    if gpu_num > 1:
        model = multi_gpu_model(model, gpus=gpu_num)

    return model


def multi_label_classification_model_2(num_classes, gpu_num=1,
                                       learning_rate=0.001):
    model = Sequential()
    model.add(Conv2D(32, (3, 3), padding="same",
                     input_shape=MultiLabelClassificationConfig.INPUT_SHAPE))
    model.add(Activation("relu"))
    model.add(BatchNormalization(axis=-1))
    model.add(MaxPooling2D(pool_size=(3, 3)))
    model.add(Dropout(0.25))

    model.add(Conv2D(64, (3, 3), padding="same"))
    model.add(Activation("relu"))
    model.add(BatchNormalization(axis=-1))
    model.add(Conv2D(64, (3, 3), padding="same"))
    model.add(Activation("relu"))
    model.add(BatchNormalization(axis=-1))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Dropout(0.25))

    model.add(Flatten())
    model.add(Dense(1024))
    model.add(Activation("relu"))
    model.add(BatchNormalization())
    model.add(Dropout(0.5))

    model.add(Dense(num_classes))
    model.add(Activation("sigmoid"))

    if gpu_num > 1:
        model = multi_gpu_model(model, gpus=gpu_num)

    return model


####################DEBLUR_GAN_MODELS####################
def generator_model():
    ngf = DeblurConfig.NGF
    n_blocks_gen = DeblurConfig.N_BLOCKS_GEN
    shape = DeblurConfig.CELL_SHAPE
    output_nc = DeblurConfig.OUTPUT_NC

    # Build generator architecture.
    # Current version : ResNet block
    inputs = Input(shape=shape)

    x = ReflectionPadding2D((3, 3))(inputs)
    x = Conv2D(filters=ngf, kernel_size=(7, 7), padding='valid')(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)

    n_downsampling = 2
    for i in range(n_downsampling):
        mult = 2 ** i
        x = Conv2D(filters=ngf * mult * 2, kernel_size=(3, 3),
                   strides=2, padding='same')(x)
        x = BatchNormalization()(x)
        x = Activation('relu')(x)

    mult = 2 ** n_downsampling
    for i in range(n_blocks_gen):
        x = res_block(x, ngf * mult, use_dropout=True)

    for i in range(n_downsampling):
        mult = 2 ** (n_downsampling - i)
        # x = Conv2DTranspose(filters=int(ngf * mult / 2), kernel_size=(3, 3), strides=2, padding='same')(x)
        x = UpSampling2D()(x)
        x = Conv2D(filters=int(ngf * mult / 2),
                   kernel_size=(3, 3), padding='same')(x)
        x = BatchNormalization()(x)
        x = Activation('relu')(x)

    x = ReflectionPadding2D((3, 3))(x)
    x = Conv2D(filters=output_nc, kernel_size=(7, 7), padding='valid')(x)
    x = Activation('tanh')(x)

    outputs = Add()([x, inputs])
    # outputs = Lambda(lambda z: K.clip(z, -1, 1))(x)
    outputs = Lambda(lambda z: z / 2)(outputs)

    model = Model(inputs=inputs, outputs=outputs, name='Generator')
    return model


def discriminator_model():
    ndf = DeblurConfig.NDF
    shape = DeblurConfig.CELL_SHAPE
    # Build discriminator architecture.
    n_layers, use_sigmoid = 3, False
    inputs = Input(shape=shape)

    x = Conv2D(filters=ndf, kernel_size=(4, 4),
               strides=2, padding='same')(inputs)
    x = LeakyReLU(0.2)(x)

    nf_mult, nf_mult_prev = 1, 1
    for n in range(n_layers):
        nf_mult_prev, nf_mult = nf_mult, min(2 ** n, 8)
        x = Conv2D(filters=ndf * nf_mult, kernel_size=(4, 4),
                   strides=2, padding='same')(x)
        x = BatchNormalization()(x)
        x = LeakyReLU(0.2)(x)

    nf_mult_prev, nf_mult = nf_mult, min(2 ** n_layers, 8)
    x = Conv2D(filters=ndf * nf_mult, kernel_size=(4, 4),
               strides=1, padding='same')(x)
    x = BatchNormalization()(x)
    x = LeakyReLU(0.2)(x)

    x = Conv2D(filters=1, kernel_size=(4, 4), strides=1, padding='same')(x)
    if use_sigmoid:
        x = Activation('sigmoid')(x)

    x = Flatten()(x)
    x = Dense(1024, activation='tanh')(x)
    x = Dense(1, activation='sigmoid')(x)

    model = Model(inputs=inputs, outputs=x, name='Discriminator')
    return model


def generator_containing_discriminator(generator, discriminator):
    image_shape = DeblurConfig.CELL_SHAPE
    inputs = Input(shape=image_shape)
    generated_image = generator(inputs)
    outputs = discriminator(generated_image)
    model = Model(inputs=inputs, outputs=outputs)
    return model


def generator_containing_discriminator_multiple_outputs(
        generator, discriminator):
    image_shape = DeblurConfig.CELL_SHAPE
    inputs = Input(shape=image_shape)
    generated_image = generator(inputs)
    outputs = discriminator(generated_image)
    model = Model(inputs=inputs, outputs=[generated_image, outputs])
    return model



import os
import cv2
import numpy as np
from keras.utils import Sequence
from keras.optimizers import Adam
from .model import unet
from .....lib.general_lib import get_data_path_type
from ....application import Application


class Unet(Application):
    def __init__(self, weights_path=None):
        super().__init__()

        self.model = self.create_model()

        if weights_path is not None:
            self.load(weights_path)

    def train(self, data_path, epochs, learning_rate=1e-4, batch_size=1):
        data_gen = TrainGenerator()
        data_generator = data_gen.flow_from_directory(data_path,
                                                      image_shape=(512, 512),
                                                      batch_size=batch_size)

        self.compile_model(learning_rate)

        self.model.fit_generator(data_generator,
                                 steps_per_epoch=len(data_generator),
                                 epochs=epochs)

    def predict(self, data_path, batch_size=1, result_folder='results'):
        if not os.path.isdir(result_folder):
            os.makedirs(result_folder)

        data_type = get_data_path_type(data_path)
        if data_type == 'image':
            filename = os.path.basename(data_path)
            img = load_image_with_proper_format(data_path, (512, 512))
            img = np.reshape(img, (1,) + img.shape)
            result = self.model.predict(img)
            result = result * 255
            cv2.imwrite(os.path.join(result_folder, filename), result[0])
        elif data_type == 'directory':
            saved_folder = os.path.join(result_folder, data_path)
            if not os.path.isdir(saved_folder):
                os.makedirs(saved_folder)

            data_gen = PredictGenerator()
            data_generator = data_gen.flow_from_directory(data_path,
                                                          image_shape=(
                                                              512, 512),
                                                          batch_size=batch_size)
            results = self.model.predict_generator(data_generator,
                                                   steps=len(data_generator))
            for i, result in enumerate(results):
                filename = data_generator.predict_order_filenames[i]
                saved_path = os.path.join(saved_folder, filename)
                result = result * 255
                cv2.imwrite(saved_path, result)

    def save(self, weights_path):
        self.model.save_weights(weights_path)

    def load(self, weights_path):
        self.model.load_weights(weights_path)

    def create_model(self):
        return unet()

    def compile_model(self, learning_rate):
        self.model.compile(optimizer=Adam(lr=learning_rate),
                           loss='binary_crossentropy',
                           metrics=['accuracy'])


def load_image_with_proper_format(image_path, image_shape):
    img = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
    img = img / 255
    img = cv2.resize(img, image_shape)
    img = np.reshape(img, img.shape + (1,))
    return img


class TrainGenerator(Sequence):
    def __init__(self, image_filenames=None, mask_filenames=None,
                 image_shape=None):
        self.image_filenames = image_filenames
        self.mask_filenames = mask_filenames
        self.image_shape = image_shape
        self.batch_size = 1

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        images_names = self.image_filenames[
            idx * self.batch_size:(idx + 1) * self.batch_size]
        masks_names = self.mask_filenames[
            idx * self.batch_size:(idx + 1) * self.batch_size]

        images = []
        masks = []
        for image_name in images_names:
            try:
                img = load_image_with_proper_format(image_name,
                                                    self.image_shape)
                images.append(img)
            except BaseException:
                pass
        for mask_name in masks_names:
            try:
                mask = load_image_with_proper_format(mask_name,
                                                     self.image_shape)
                masks.append(mask)
            except BaseException:
                pass

        return np.array(images), np.array(masks)

    def flow_from_directory(self, data_path, image_shape, batch_size=1):
        image_folder_path = os.path.join(data_path, 'image')
        mask_folder_path = os.path.join(data_path, 'mask')

        images = os.listdir(image_folder_path)
        masks = os.listdir(mask_folder_path)

        self.image_filenames = [
            f'{image_folder_path}/{image}' for image in images
        ]
        self.mask_filenames = [
            f'{mask_folder_path}/{mask}' for mask in masks
        ]
        self.batch_size = batch_size
        self.image_shape = image_shape

        return self


class PredictGenerator(Sequence):
    def __init__(self, image_filenames=None, image_shape=None):
        self.image_filenames = image_filenames
        self.image_shape = image_shape
        self.batch_size = 1
        self.predict_order_filenames = []

    def __len__(self):
        return int(np.ceil(len(self.image_filenames) / float(self.batch_size)))

    def __getitem__(self, idx):
        images_names = self.image_filenames[
            idx * self.batch_size:(idx + 1) * self.batch_size]

        images = []
        for image_name in images_names:
            try:
                _, only_file_name = os.path.split(image_name)
                image = load_image_with_proper_format(image_name,
                                                      self.image_shape)
                images.append(image)
                self.predict_order_filenames.append(only_file_name)
            except BaseException:
                pass

        return np.array(images)

    def flow_from_directory(self, data_path, image_shape, batch_size=1):
        images = os.listdir(data_path)

        self.image_filenames = [
            f'{data_path}/{image}' for image in images
        ]
        self.batch_size = batch_size
        self.image_shape = image_shape

        return self



import numpy as np


class DeepblockLogConfig:
    NAME = "DEEPBLOCK_LOG"
    LOG_TYPE = "logType"
    PREDICT_START = "predictStarted"
    PREDICT_END = "predictFinished"
    SAVE_START = "startSavingModel"
    SAVE_END = "finishSavingModel"
    RESULT = "result"
    SUCCESS = "success"
    TRAIN_START = "trainStarted"
    TRAIN_END = "trainFinished"


class ImageClassificationConfig:
    NAME = "IMAGE_CLASSIFICATION"
    INPUT_SHAPE = (224, 224, 3)


class MultiLabelClassificationConfig:
    NAME = "MULTI_LABEL_CLASSIFICATION"
    INPUT_SHAPE = (224, 224, 3)


class DeblurConfig:
    NAME = 'CONFIG_DEBLUR'
    RESULT_FOLDER = "results/deblur"
    NGF = 64
    NDF = 64
    OUTPUT_NC = 3
    N_BLOCKS_GEN = 9
    CELL_SHAPE = (512, 512, 3)


class ObjectDetectionConfig:
    NAME = "OBJECT_DETECTION"
    RESULT_FOLDER = "results/object_detection"
    INPUT_SHAPE = (416, 416)
    RECTANGLE_COLOR = (153, 204, 153)
    TEXT_COLOR = (51, 204, 204)
    ANCHORS = np.array([10.0, 13.0,
                        16.0, 30.0,
                        33.0, 23.0,
                        30.0, 61.0,
                        62.0, 45.0,
                        59.0, 119.0,
                        116.0, 90.0,
                        156.0, 198.0,
                        373.0, 326.0]).reshape(-1, 2)


class MaskRCNNConfig:
    NAME = 'MASK_RCNN'
    RESULT_FOLDER = "results/segmentation"
    GPU_COUNT = 1
    IMAGES_PER_GPU = 2
    STEPS_PER_EPOCH = 1000
    VALIDATION_STEPS = 50
    BACKBONE = "resnet101"
    COMPUTE_BACKBONE_SHAPE = None
    BACKBONE_STRIDES = [4, 8, 16, 32, 64]
    FPN_CLASSIF_FC_LAYERS_SIZE = 1024
    TOP_DOWN_PYRAMID_SIZE = 256
    NUM_CLASSES = 1  # Override in sub-classes
    RPN_ANCHOR_SCALES = (32, 64, 128, 256, 512)
    RPN_ANCHOR_RATIOS = [0.5, 1, 2]
    RPN_ANCHOR_STRIDE = 1
    RPN_NMS_THRESHOLD = 0.7
    RPN_TRAIN_ANCHORS_PER_IMAGE = 256
    PRE_NMS_LIMIT = 6000
    POST_NMS_ROIS_TRAINING = 2000
    POST_NMS_ROIS_INFERENCE = 1000
    USE_MINI_MASK = True
    MINI_MASK_SHAPE = (56, 56)  # (height, width) of the mini-mask
    IMAGE_RESIZE_MODE = "square"
    IMAGE_MIN_DIM = 800
    IMAGE_MAX_DIM = 1024
    IMAGE_MIN_SCALE = 0
    IMAGE_CHANNEL_COUNT = 3
    MEAN_PIXEL = np.array([123.7, 116.8, 103.9])
    TRAIN_ROIS_PER_IMAGE = 200
    ROI_POSITIVE_RATIO = 0.33
    POOL_SIZE = 7
    MASK_POOL_SIZE = 14
    MASK_SHAPE = [28, 28]
    MAX_GT_INSTANCES = 100
    RPN_BBOX_STD_DEV = np.array([0.1, 0.1, 0.2, 0.2])
    BBOX_STD_DEV = np.array([0.1, 0.1, 0.2, 0.2])
    DETECTION_MAX_INSTANCES = 100
    DETECTION_MIN_CONFIDENCE = 0.7
    DETECTION_NMS_THRESHOLD = 0.3
    LEARNING_RATE = 0.001
    LEARNING_MOMENTUM = 0.9
    WEIGHT_DECAY = 0.0001
    LOSS_WEIGHTS = {
        "rpn_class_loss": 1.,
        "rpn_bbox_loss": 1.,
        "mrcnn_class_loss": 1.,
        "mrcnn_bbox_loss": 1.,
        "mrcnn_mask_loss": 1.
    }
    USE_RPN_ROIS = True
    TRAIN_BN = False  # Defaulting to False since batch size is often small

    # Gradient norm clipping
    GRADIENT_CLIP_NORM = 5.0

    def __init__(self):
        """Set values of computed attributes."""
        # Effective batch size
        self.BATCH_SIZE = self.IMAGES_PER_GPU * self.GPU_COUNT

        # Input image size
        if self.IMAGE_RESIZE_MODE == "crop":
            self.IMAGE_SHAPE = np.array([self.IMAGE_MIN_DIM, self.IMAGE_MIN_DIM,
                                         self.IMAGE_CHANNEL_COUNT])
        else:
            self.IMAGE_SHAPE = np.array([self.IMAGE_MAX_DIM, self.IMAGE_MAX_DIM,
                                         self.IMAGE_CHANNEL_COUNT])

        # Image meta data length
        # See compose_image_meta() for details
        self.IMAGE_META_SIZE = 1 + 3 + 3 + 4 + 1 + self.NUM_CLASSES

    def display(self):
        """Display Configuration values."""
        print("\nConfigurations:")
        for a in dir(self):
            if not a.startswith("__") and not callable(getattr(self, a)):
                print("{:30} {}".format(a, getattr(self, a)))
        print("\n")


class SuperResolutionConfig:
    NAME = 'SUPER_RESOLUTION'
    RESULT_FOLDER = "results/super_resolution"
    IMAGES_FOLDER = "cropped_images"
    LR_FOLDER = "lr"
    SR_FOLDER = "sr"
    NUM_FILTERS = 64
    NUM_RES_BLOCKS = 16
    IMAGE_WIDTH = 8
    IMAGE_HEIGHT = 8
    IMAGE_CHANNELS = 3
    SCALE = 2



import platform


class Application:

    def __init__(self):
        try:
            self.os_type = platform.system()
        except Exception as e:
            print("Application initialize failed")
            raise e
        # default setting for using 1gpu and gpu_ids for 0
        self.gpu_num = 1
        self.gpu_ids = [0]
        self.deepblock_log = False

    def use_specific_gpus(self, gpu_ids):
        self.gpu_ids = gpu_ids
        self.gpu_num = len(gpu_ids)

    def use_deepblock_site_log(self, use_or_not=False):
        self.deepblock_log = use_or_not



"""This module is a custom module designed for general use in the Omnis project.
"""
import os

import numpy as np


def reverse_dict(input_dict):
    """This function returns a dictionary with input_dict's key, value is
    reversed. If input_dict has same value among different keys, some keys
    may be omitted.

    Arguments:
        input_dict {dict} -- Dictionary to reverse.

    Returns:
        [dict] -- Reversed dictionary.
    """
    return {value: key for key, value in input_dict.items()}


def get_n_largest_index_and_probs(probs_array, n):
    """This function returns array of top n index of probs

    Arguments:
        probs_array {array of array} -- Array of probs array
        n {int} -- top n

    Returns:
        [array of array] -- Array of Top n index array
        [array of array] -- Array of Top n probs array
    """
    if n > probs_array[0].size:
        raise ValueError("n should be smaller than classes number")

    indexes = np.argsort(-probs_array)[:, :n]
    sorted_probs = np.zeros((len(probs_array), n))
    for i, prob in enumerate(probs_array):
        for j in range(0, n):
            sorted_probs[i][j] = (int(prob[indexes[i][j]] * 10000)) / 10000

    return indexes, sorted_probs


def get_data_path_type(data_path):
    """This function returns type of data_path

    Arguments:
        data_path {string} -- name of data_path

    Returns:
        [string] -- return directory, image, csv decided by data_path type
    """
    if os.path.isdir(data_path):
        return "directory"

    _, file_extension = os.path.splitext(data_path)

    image_extension_list = ['.jpg', '.jpeg', '.png', '.bmp', '.ppm', '.tif',
                            '.tiff']

    if file_extension in image_extension_list:
        return "image"

    if file_extension == ".csv":
        return "csv"

    raise ValueError("We didn't provide %s type" % file_extension)


def get_area_by_polygons(all_points_x, all_points_y):
    """This function returns the area calculated by polygons.
    For more detailed explanation, refer to the Wikipedia article:
    https://en.wikipedia.org/wiki/Shoelace_formula

    Arguments:
        all_points_x {numpy array} -- np of x
        all_points_y {numpy array} -- np of y

    Returns:
        [float] -- return the area of polygons
    """
    return 0.5 * np.abs(
        np.dot(all_points_x, np.roll(all_points_y, 1))
        - np.dot(all_points_y, np.roll(all_points_x, 1))
    )


def divide_train_val_dataset(data_path):
    """This function returns lists of train data paths and validation
    data paths, with ratio of 9:1.

    Arguments:
        data_path {str} -- data path

    Returns:
        train {list} -- list of paths to train data
        val {list} -- list of paths to validation data
    """
    images = os.listdir(data_path)
    train = []
    val = []
    for i, image in enumerate(images):
        if i % 10 == 0:
            val.append(os.path.join(data_path, image))
        else:
            train.append(os.path.join(data_path, image))
    return train, val



# -*- coding: utf-8 -*-
"""This module is a utility module for image preprocessing.
"""
import math
import os
import sys

import cv2
import numpy

IMAGE_EXTENSIONS = ('png', 'jpg', 'jpeg',)


def reshape_data(data_array, input_shape, cv_interpolation=cv2.INTER_LINEAR):
    """Changes an array of images to fit with input shape.
    Note that this method casts and rescales data to a range of [0, 1].

    Arguments:
        data_array {ndarray} -- An array of images.
        input_shape -- Shape of input

    Keyword Arguments:
        cv_interpolation {int} -- cv2 interpolation. (default: {cv2.INTER_LINEAR})

    Returns:
        [ndarray] -- Reshaped data array.
    """
    if data_array.shape[1:] != input_shape:
        reshaped_list = list()
        for i in range(data_array.shape[0]):
            resized_img = cv2.resize(
                data_array[i],
                (input_shape[1], input_shape[2]),
                interpolation=cv_interpolation)
            reshaped_list.append(resized_img)
        data_array = numpy.asarray(reshaped_list)
        reshaped_list = []  # empty list after use
    data_array = data_array.astype('float32')
    data_array /= 255
    return data_array


def resize_image(img, target_size, cv_interpolation=cv2.INTER_LINEAR):
    """Resizes image with specified target size and casts to float32.

    Arguments:
        img {ndarray} -- An opencv image array
        target_size {tuple} -- A tuple of two natural numbers. Two numbers represent the width and height, respectively.

    Keyword Arguments:
        cv_interpolation {int} -- cv2 interpolation. (default: {cv2.INTER_LINEAR})

    Returns:
        [ndarray] -- Resized image array.
    """
    resized_img = cv2.resize(img, target_size, interpolation=cv_interpolation)
    resized_img = resized_img.astype('float32')
    return resized_img


def is_image_file(filename):
    """this function checks a file extension.

    Arguments:
        filename {str} -- A path of file.

    Returns:
        [bool] -- If an extension of the file is an image, returns True. Otherwise False.
    """
    return '.' in filename and filename.rsplit('.', 1)[1] in IMAGE_EXTENSIONS


def get_current_directory_name(file_path):
    """This function returns a parent directory name of the file at the file_path.

    Arguments:
        file_path {str} -- A path of the file.

    Raises:
        ValueError -- If the file is not in a directory, then this function raises a ValueError.

    Returns:
        [str or unicode] -- A parent directory name.(The same as a label of the file.)
    """
    splitted_filename = None
    if sys.platform == "win32":
        splitted_filename = file_path.split('\\')
    else:
        splitted_filename = file_path.split('/')
    if len(splitted_filename) < 2:
        raise ValueError(
            'No parent directory found. You should put your image files inside '
            'a directory name which is the same as a label of them.'
        )

    return splitted_filename[-2]


def append_elem(path, img_list, label_list, target_size):
    """This function recursively iterates over files and directories.
    While iterating, if an image file found,
    this function appends an image as an array format to img_list and appends a label to label_list

    Arguments:
        path {str} -- A path of a current file or a director.
        img_list {list} -- A list of images.
        label_list {list} -- A list of labels.
        target_size {tuple} -- A tuple which specifies a target size of image array.(Image files are resized to target size.)
    """
    if os.path.isdir(path):
        child_list = os.listdir(path)
        for child in child_list:
            child_path = os.path.join(path, child)
            append_elem(child_path, img_list, label_list, target_size)
    else:
        if is_image_file(path):
            img = cv2.imread(path)
            resized_image = resize_image(img, target_size)
            label = get_current_directory_name(path)
            img_list.append(resized_image)
            label_list.append(label)


def directory_images_to_arrays(directory_path, target_size):
    """Recursively iterates over child files and directories of directory_path.
    Image files will be converted to an array of images with rescaled value.
    Parent directory names will be converted to array_of_labels.

    Arguments:
        directory_path {str} -- A root path of image files and directories.
        target_size {tuple} -- A tuple which specifies a target size of image
        array.(Image files are resized to target size.)

    Returns:
        array_of_images, array_of_labels [(ndarray, ndarray)] --
            array_of_images is a rescaled array of resized images.
            array_of_labels is a label array of images.
            You can find a label of an image at the same index of array_of_labels.
    """
    img_list = list()
    label_list = list()
    append_elem(directory_path, img_list, label_list, target_size)
    array_of_images = numpy.asarray(img_list)
    array_of_images /= 255
    array_of_labels = numpy.asarray(label_list)
    return array_of_images, array_of_labels


def get_image_names_in_directory(directory_path):
    """Iterates over files in directory_path.
    Return image file path lists of below directory_path.

    Arguments:
        directory_path {str} -- A root path of image files.
        target_size {tuple} -- A tuple which specifies a target size of image array.(Image files are resized to target size.)

    Returns:
        [str list] --
            list of image file paths inside given directory_path.
    """
    files = sorted(os.listdir(directory_path))
    return [os.path.join(directory_path, f) for f in files if is_image_file(f)]


def merge_image(img_list, images_area_index):
    """Merge images in img_list using images_area_index
    void area will be filled with 0 value
    Return image numpy array which merging given images

    Arguments:
        img_list {str list} -- Filenames of image files
        images_area_index {int[4] list} -- x1,y1,x2,y2 list of images

    Returns:
        [nparray] --
            array of merged image
    """
    boundary_xs = set()
    boundary_ys = set()

    for images_area in images_area_index:
        boundary_xs.add(images_area[2])
        boundary_ys.add(images_area[3])

    boundary_xs = list(boundary_xs)
    boundary_ys = list(boundary_ys)

    BG = numpy.zeros((max(boundary_ys), max(boundary_xs), 3), numpy.uint8)

    for index, image in enumerate(img_list):
        image_original_size = (
            images_area_index[index][2] -
            images_area_index[index][0],
            images_area_index[index][3] -
            images_area_index[index][1])
        image_after_resize = resize_image(image, image_original_size)

        # image_after_resize_blur = blur_boundary_area(image_after_resize, [2, image_original_size[0]-2], [2, image_original_size[1]-2], 2)
        BG[
            images_area_index[index][1]:images_area_index[index][3],
            images_area_index[index][0]:images_area_index[index][2]
        ] = image_after_resize

    # boundary_xs.remove(max(boundary_xs))
    # boundary_ys.remove(max(boundary_ys))
    # BG = blur_boundary_area(BG, boundary_xs, boundary_ys, 2)
    return BG


def blur_boundary_area(img, boundary_xs, boundary_ys, pixel_range):
    for y in boundary_ys:
        y_line = img[y - pixel_range:y + pixel_range, :]
        blur = cv2.GaussianBlur(y_line,
                                (2 * pixel_range + 1, 2 * pixel_range + 1), 0)
        img[y - pixel_range:y + pixel_range] = blur

    for x in boundary_xs:
        x_line = img[:, x - pixel_range:x + pixel_range]
        blur = cv2.GaussianBlur(x_line,
                                (2 * pixel_range + 1, 2 * pixel_range + 1), 0)
        img[:, x - pixel_range:x + pixel_range] = blur

    return img


def split_one_image(img, split_width, split_height):
    """Split images in img by split_width and split_height
    Return list of split image array and area_index for each image array

    Arguments:
        img {array} -- array of image files
        split_width {int} -- width you want to split in this size
        split_height {int} -- height you want to split in this size

    Returns:
        [list of array] -- list of image array
        [list] -- list of (x1, y1, x2, y2)
    """
    width, height = img.shape[1], img.shape[0]
    x_count = math.ceil(width / split_width)
    y_count = math.ceil(height / split_height)

    split_image_list = list()
    area_index = list()
    for i in range(0, x_count):
        for j in range(0, y_count):
            x_start = i * split_width
            x_end = min((i + 1) * split_width, width)
            y_start = j * split_height
            y_end = min((j + 1) * split_height, height)
            area = (x_start, y_start, x_end, y_end)
            cropped_img = img[y_start:y_end, x_start:x_end]
            split_image_list.append(cropped_img)
            area_index.append(area)

    return split_image_list, area_index


def save_image(image, file_path):
    directory = os.path.dirname(file_path)

    if not os.path.exists(directory):
        os.makedirs(directory)

    cv2.imwrite(file_path, image)



from keras.applications.vgg16 import VGG16
import keras.backend as K
from keras.models import Model
import numpy as np


def l1_loss(y_true, y_pred):
    return K.mean(K.abs(y_pred - y_true))


def perceptual_loss_100(y_true, y_pred):
    return 100 * perceptual_loss(y_true, y_pred)


def perceptual_loss(y_true, y_pred):
    vgg = VGG16(include_top=False, weights='imagenet',
                input_shape=(256, 256, 3))
    loss_model = Model(
        inputs=vgg.input, outputs=vgg.get_layer('block3_conv3').output)
    loss_model.trainable = False
    return K.mean(K.square(loss_model(y_true) - loss_model(y_pred)))


def wasserstein_loss(y_true, y_pred):
    return K.mean(y_true * y_pred)


def gradient_penalty_loss(y_true, y_pred, averaged_samples):
    # pylint: disable=unused-argument
    gradients = K.gradients(y_pred, averaged_samples)[0]
    gradients_sqr = K.square(gradients)
    gradients_sqr_sum = K.sum(gradients_sqr,
                              axis=np.arange(1, len(gradients_sqr.shape)))

    gradient_l2_norm = K.sqrt(gradients_sqr_sum)
    gradient_penalty = K.square(1 - gradient_l2_norm)

    return K.mean(gradient_penalty)


def psnr_loss(y_true, y_pred):
    return -10. * K.log(K.mean(K.square(y_pred - y_true))) / K.log(10.)



from collections import OrderedDict
import json
import os

import cv2
import numpy as np
import skimage.draw
import skimage.draw
import skimage.io
import skimage.io

from ...application import Application
from ....lib.config_lib import DeepblockLogConfig, MaskRCNNConfig
from ....lib.custom_callback import ImageClassificationCallback
from ....lib.general_lib import divide_train_val_dataset, get_data_path_type
# Import Mask RCNN
from ....lib.mask_rcnn import model as model_lib, utils, visualize


class ImageSegmentation(Application):
    def __init__(self, class_list_path=None, weights_path=None, backbone=None,
                 default_weights_path='coco.h5'):
        super().__init__()

        self.config = CustomConfig()
        self.weights_path = weights_path
        self.action = None
        self.default_weights_path = default_weights_path
        self.model = None

        if backbone is not None:
            self.config.BACKBONE = backbone
        if weights_path is not None:
            if class_list_path is None:
                raise ValueError(
                    "you should write class list and weights_path both")
            else:
                self.load(class_list_path, weights_path)

    def train(self,
              data_path,
              annotation_path, learning_rate=0.01,
              epochs=10,
              layers='heads',
              callbacks=[],
              verbose=0):
        dataset_train, dataset_val = \
            self.make_train_valid_dataset_and_set_class_list(data_path,
                                                             annotation_path)

        if self.action != "train":
            self.create_model(mode="training")
            self.load_weights_for_train()
            self.action = "train"

        if self.deepblock_log:
            deepblock_log_json = OrderedDict()
            deepblock_log_json[
                DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.TRAIN_START
            print(json.dumps(deepblock_log_json, ensure_ascii=False))

        if self.deepblock_log:
            callbacks += [ImageClassificationCallback(
                total_step=self.config.STEPS_PER_EPOCH, total_epoch=epochs)]
        self.model.train(dataset_train, dataset_val,
                         learning_rate=learning_rate,
                         epochs=epochs,
                         layers=layers,
                         custom_callbacks=callbacks,
                         verbose=verbose)

        if self.deepblock_log:
            deepblock_log_json = OrderedDict()
            deepblock_log_json[DeepblockLogConfig.LOG_TYPE] = \
                DeepblockLogConfig.TRAIN_END
            print(json.dumps(deepblock_log_json, ensure_ascii=False))

    def predict(
            self,
            data_path=None,
            verbose=0,
            result_path=MaskRCNNConfig.RESULT_FOLDER):
        if self.action != "predict":
            self.create_model(mode="inference")
            self.load_weights_for_predict()
            self.action = "predict"

        get_image_from = get_data_path_type(data_path)

        if self.deepblock_log:
            deepblock_log_json = OrderedDict()
            deepblock_log_json[
                DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.PREDICT_START
            print(json.dumps(deepblock_log_json, ensure_ascii=False))

        if get_image_from == "directory":
            for image_file in os.listdir(data_path):
                self.save_after_detect_one_image(
                    os.path.join(data_path, image_file), result_path)
        elif get_image_from == "image":
            self.save_after_detect_one_image(data_path, result_path)

        if self.deepblock_log:
            deepblock_log_json = OrderedDict()
            deepblock_log_json[
                DeepblockLogConfig.LOG_TYPE] = DeepblockLogConfig.PREDICT_END
            print(json.dumps(deepblock_log_json, ensure_ascii=False))

    def save(self, class_list_path, weights_path):
        self.model.keras_model.save_weights(weights_path)
        with open(class_list_path, 'w') as f:
            for i in range(self.config.NUM_CLASSES):
                f.write("%s" % self.config.CLASS_NAMES[i])
                i += 1
                if i != self.config.NUM_CLASSES:
                    f.write("\n")

    def load(self, class_list_path=None, weights_path=None):
        if isinstance(class_list_path, type(None)):
            raise ValueError("Class names should be given")
        if isinstance(weights_path, type(None)):
            raise ValueError("Weights path should be given")

        class_file = open(class_list_path, 'r')
        class_objects = class_file.read()
        class_file.close()
        class_list = class_objects.split('\n')
        # remove last class_list
        self.set_config(class_list=class_list)
        self.weights_path = weights_path

    def create_weights_folder(self, weights_path):
        weights_folder = os.path.dirname(weights_path)

        if weights_folder != '':
            if not os.path.exists(weights_folder):
                os.makedirs(weights_folder)

    def load_weights_for_train(self):
        self.create_weights_folder(self.default_weights_path)

        if isinstance(self.weights_path, type(None)):
            self.weights_path = self.default_weights_path
            if not os.path.exists(self.default_weights_path):
                utils.download_trained_weights(self.weights_path)
        self.model.load_weights(self.weights_path, by_name=True, exclude=[
            "mrcnn_class_logits", "mrcnn_bbox_fc",
            "mrcnn_bbox", "mrcnn_mask"])

    def load_weights_for_predict(self):
        self.create_weights_folder(self.default_weights_path)

        if isinstance(self.weights_path, type(None)):
            if not os.path.exists(self.default_weights_path):
                utils.download_trained_weights(self.weights_path)
        self.model.load_weights(self.weights_path, by_name=True)

    def create_model(self, mode="training"):
        self.config.GPU_COUNT = self.gpu_num
        self.config.BATCH_SIZE = self.config.GPU_COUNT * \
                                 self.config.IMAGES_PER_GPU
        self.model = model_lib.MaskRCNN(
            mode=mode, config=self.config, model_dir='.')
        self.model.config.display()

    def fill_true_by_multiple_masks(self, width, height, masks):
        total_board = np.zeros((height, width), dtype=bool)
        for i in range(masks.shape[-1]):
            mask = masks[:, :, i]
            total_board = np.bitwise_or(total_board, mask)
        return total_board

    def fill_true_by_polygons(self, width, height, file_polygons):
        total_board = np.zeros((height, width))
        for polygon in file_polygons:
            one_polygon = polygon
            all_points_x = np.asarray(one_polygon['all_points_x'])
            all_points_y = np.asarray(one_polygon['all_points_y'])
            pair_x_y = np.array(zip(all_points_x, all_points_y))

            total_board = cv2.fillPoly(total_board, pts=[pair_x_y], color=1)
        return total_board.astype(bool)

    def make_train_valid_dataset_and_set_class_list(self, data_path,
                                                    annotation_path):
        self.train_dataset, self.val_dataset = divide_train_val_dataset(
            data_path)

        dataset_train = ImageDataset()
        dataset_train.load_class(self.train_dataset, annotation_path)
        dataset_train.prepare()
        print("Images: {}\nClasses: {}".format(
            len(dataset_train.image_ids), dataset_train.class_names))

        dataset_val = ImageDataset()
        dataset_val.load_class(self.val_dataset, annotation_path)
        dataset_val.prepare()
        print("Images: {}\nClasses: {}".format(
            len(dataset_val.image_ids), dataset_val.class_names))

        assert sorted(dataset_train.class_names) == sorted(
            dataset_val.class_names), \
            "train classes have to be same with val classes"
        self.set_config(class_list=dataset_train.class_names)

        return dataset_train, dataset_val

    def detect_one_image(self, file_path, verbose=1):
        predict_image = skimage.io.imread(file_path)
        results = self.model.detect([predict_image], verbose=verbose)
        r = results[0]
        image_data = [predict_image]

        return r, image_data

    def save_after_detect_one_image(self, file_path, result_path):
        _, filename = os.path.split(file_path)
        r, image_data = self.detect_one_image(file_path)

        if not os.path.isdir(result_path):
            os.makedirs(result_path)

        plt = self.mask(image_data=image_data, r=r)
        plt.savefig(os.path.join(result_path, filename))

    def set_config(self, class_list, detection_threshold=0.8,
                   backbone='resnet50'):
        self.config.CLASS_NAMES = class_list
        self.config.NUM_CLASSES = len(class_list)
        self.config.IMAGE_META_SIZE = 12 + self.config.NUM_CLASSES
        self.config.DETECTION_MIN_CONFIDENCE = detection_threshold
        self.config.BACKBONE = backbone

    def mask(self, image_data, r, show_mask=True):
        plt, _ = visualize.display_instances(image_data[0], r['rois'],
                                             r['masks'], r['class_ids'],
                                             self.config.CLASS_NAMES,
                                             r['scores'],
                                             show_mask=show_mask)
        return plt


class ImageDataset(utils.Dataset):
    def image_reference(self, image_id):
        """Return the path of the image."""
        info = self.image_info[image_id]
        if info["source"] == "source":
            return info["path"]
        else:
            return super().image_reference(image_id)

    def load_mask(self, image_id):
        """Generate instance masks for an image.
       Returns:
        masks: A bool array of shape [height, width, instance count] with
            one mask per instance.
        class_ids: a 1D array of class IDs of the instance masks.
        """
        image_info = self.image_info[image_id]
        if image_info["source"] != "source":
            return super().load_mask(image_id)

        info = self.image_info[image_id]
        mask = np.zeros([info["height"], info["width"], len(info["polygons"])],
                        dtype=np.uint8)
        for i, p in enumerate(info["polygons"]):
            rr, cc = skimage.draw.polygon(p['all_points_y'], p['all_points_x'])
            mask[rr, cc, i] = 1
        if info["class_ids"]:
            class_ids = np.array(info["class_ids"], dtype=np.int32)
            return mask.astype(np.bool), class_ids
        else:
            return super().load_mask(image_id)

    def load_class(self, dataset, annotation_path):
        """Load annotations of the dataset and add classes.
        dataset: List of the dataset.
        """
        dataset_dir = os.path.dirname(dataset[0])

        coco_json = json.load(
            open(annotation_path))
        classes = []
        for c in coco_json['categories']:
            if c['name'] not in classes:
                classes.append(c['name'])

        class_list = sorted(classes)
        j = 1
        for class_name in class_list:
            self.add_class("source", j, class_name)
            j += 1

        image_id = -1
        polygons = []
        class_ids = []
        j = 0
        for a in sorted(coco_json['annotations'],
                        key=lambda ann: ann['imageId']):
            annotations = {'name': 'polygon'}
            x = []
            y = []
            for i, coordinate in enumerate(a['segmentation']):
                if i % 2 == 0:
                    x.append(coordinate)
                else:
                    y.append(coordinate)
            annotations['all_points_x'] = x
            annotations['all_points_y'] = y
            cl = classes[a['categoryId'] - 1]
            if image_id == a['imageId']:
                polygons.append(annotations)
                class_ids.append(class_list.index(cl) + 1)
            else:
                if j != 0:
                    filename = coco_json['images'][j - 1]['fileName']
                    image_path = os.path.join(dataset_dir, filename)
                    image = skimage.io.imread(image_path)
                    height, width = image.shape[:2]
                    self.add_image(
                        "source",
                        image_id=filename,  # use file name as a unique image id
                        path=image_path,
                        width=width, height=height,
                        polygons=polygons,
                        class_ids=class_ids)
                    class_ids = []
                class_ids.append(class_list.index(cl) + 1)
                image_id = a['imageId']
                polygons = [annotations]
                j += 1
        filename = coco_json['images'][j - 1]['fileName']
        image_path = os.path.join(dataset_dir, filename)
        image = skimage.io.imread(image_path)
        height, width = image.shape[:2]
        self.add_image(
            "source",
            image_id=filename,  # use file name as a unique image id
            path=image_path,
            width=width, height=height,
            polygons=polygons,
            class_ids=class_ids)


############################################################
#  Configurations
############################################################


class CustomConfig(MaskRCNNConfig):
    """Configuration for training on the custom dataset.
    Derives from the base Config class and overrides some values.
    """
    # Give the configuration a recognizable name
    # TODO: have to change Name of config for log etc.
    NAME = "CUSTOM_M_RCNN"

    # Number of classes (including background)
    NUM_CLASSES = 1  # Background + classes

    # Number of training steps per epoch
    STEPS_PER_EPOCH = 50

    # Skip detections with < 60% confidence
    DETECTION_MIN_CONFIDENCE = 0.8

    # Names of classes
    CLASS_NAMES = []

    IMAGE_META_SIZE = 13

    BATCH_SIZE = 1

    # BACKBONE should be resnet 50 or resnet 101
    BACKBONE = "resnet50"

    # params to reduce memory
    TRAIN_ROIS_PER_IMAGE = 100
    IMAGES_PER_GPU = 1
    MAX_GT_INSTANCES = 30



import tensorflow as tf
from keras.backend import (
    common,
    image_data_format,
)
from keras.engine import InputSpec
from keras.engine.topology import Layer
from keras.layers import (
    Activation,
    BatchNormalization,
    Conv2D,
)
from keras.layers.core import Dropout
from keras.layers.merge import Add
from keras.utils import conv_utils


# DEBLUR_GAN
def res_block(tensor_in, filters, kernel_size=(3, 3),
              strides=(1, 1), use_dropout=False):
    """
    Instanciate a Keras Resnet Block using sequential API.

    :param tensor_in: Input tensor
    :param filters: Number of filters to use
    :param kernel_size: Shape of the kernel for the convolution
    :param strides: Shape of the strides for the convolution
    :param use_dropout: Boolean value to determine the use of dropout
    :return: Keras Model
    """
    x = ReflectionPadding2D((1, 1))(tensor_in)
    x = Conv2D(filters=filters,
               kernel_size=kernel_size,
               strides=strides)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)

    if use_dropout:
        x = Dropout(0.5)(x)

    x = ReflectionPadding2D((1, 1))(x)
    x = Conv2D(filters=filters,
               kernel_size=kernel_size,
               strides=strides)(x)
    x = BatchNormalization()(x)

    merged = Add()([tensor_in, x])
    return merged


def spatial_reflection_2d_padding(
        x, padding=((1, 1), (1, 1)), data_format=None):
    """
    Pad the 2nd and 3rd dimensions of a 4D tensor.

    :param x: Input tensor
    :param padding: Shape of padding to use
    :param data_format: Tensorflow vs Theano convention ('channels_last', 'channels_first')
    :return: Tensorflow tensor
    """
    assert len(padding) == 2
    assert len(padding[0]) == 2
    assert len(padding[1]) == 2
    if data_format is None:
        data_format = image_data_format()
    if data_format not in {'channels_first', 'channels_last'}:
        raise ValueError('Unknown data_format ' + str(data_format))

    if data_format == 'channels_first':
        pattern = [[0, 0],
                   [0, 0],
                   list(padding[0]),
                   list(padding[1])]
    else:
        pattern = [[0, 0],
                   list(padding[0]), list(padding[1]),
                   [0, 0]]
    return tf.pad(x, pattern, "REFLECT")


class ReflectionPadding2D(Layer):
    """Reflection-padding layer for 2D input (e.g. picture).
    This layer can add rows and columns or zeros
    at the top, bottom, left and right side of an image tensor.
    # Arguments
        padding: int, or tuple of 2 ints, or tuple of 2 tuples of 2 ints.
            - If int: the same symmetric padding
                is applied to width and height.
            - If tuple of 2 ints:
                interpreted as two different
                symmetric padding values for height and width:
                `(symmetric_height_pad, symmetric_width_pad)`.
            - If tuple of 2 tuples of 2 ints:
                interpreted as
                `((top_pad, bottom_pad), (left_pad, right_pad))`
        data_format: A string,
            one of `channels_last` (default) or `channels_first`.
            The ordering of the dimensions in the inputs.
            `channels_last` corresponds to inputs with shape
            `(batch, height, width, channels)` while `channels_first`
            corresponds to inputs with shape
            `(batch, channels, height, width)`.
            It defaults to the `image_data_format` value found in your
            Keras config file at `~/.keras/keras.json`.
            If you never set it, then it will be "channels_last".
    # Input shape
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, rows, cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, rows, cols)`
    # Output shape
        4D tensor with shape:
        - If `data_format` is `"channels_last"`:
            `(batch, padded_rows, padded_cols, channels)`
        - If `data_format` is `"channels_first"`:
            `(batch, channels, padded_rows, padded_cols)`
    """

    def __init__(self,
                 padding=(1, 1),
                 data_format=None,
                 **kwargs):
        super(ReflectionPadding2D, self).__init__(**kwargs)
        self.data_format = common.normalize_data_format(data_format)
        if isinstance(padding, int):
            self.padding = ((padding, padding), (padding, padding))
        elif hasattr(padding, '__len__'):
            if len(padding) != 2:
                raise ValueError('`padding` should have two elements. '
                                 'Found: ' + str(padding))
            height_padding = conv_utils.normalize_tuple(padding[0], 2,
                                                        '1st entry of padding')
            width_padding = conv_utils.normalize_tuple(padding[1], 2,
                                                       '2nd entry of padding')
            self.padding = (height_padding, width_padding)
        else:
            raise ValueError('`padding` should be either an int, '
                             'a tuple of 2 ints '
                             '(symmetric_height_pad, symmetric_width_pad), '
                             'or a tuple of 2 tuples of 2 ints '
                             '((top_pad, bottom_pad), (left_pad, right_pad)). '
                             'Found: ' + str(padding))
        self.input_spec = InputSpec(ndim=4)

    def compute_output_shape(self, input_shape):
        if self.data_format == 'channels_first':
            if input_shape[2] is not None:
                rows = input_shape[2] + self.padding[0][0] + self.padding[0][1]
            else:
                rows = None
            if input_shape[3] is not None:
                cols = input_shape[3] + self.padding[1][0] + self.padding[1][1]
            else:
                cols = None
            return (input_shape[0],
                    input_shape[1],
                    rows,
                    cols)

        # self.data_format == 'channels_last'
        if input_shape[1] is not None:
            rows = input_shape[1] + self.padding[0][0] + self.padding[0][1]
        else:
            rows = None
        if input_shape[2] is not None:
            cols = input_shape[2] + self.padding[1][0] + self.padding[1][1]
        else:
            cols = None
        return (input_shape[0],
                rows,
                cols,
                input_shape[3])

    def call(self, inputs, **kwargs):
        return spatial_reflection_2d_padding(inputs,
                                             padding=self.padding,
                                             data_format=self.data_format)

    def get_config(self):
        config = {
            'padding': self.padding,
            'data_format': self.data_format
        }
        base_config = super(ReflectionPadding2D, self).get_config()
        return {**config, **base_config}



import os
import time

import cv2
import keras
from keras import backend as K
from keras.preprocessing.image import img_to_array, load_img
import numpy as np
from scipy.optimize import fmin_l_bfgs_b
import skimage

from ...application import Application


class StyleTransfer(Application):
    def __init__(self):
        super().__init__()

        self.loss_value = None
        self.grad_values = None
        self.get_content_image_from = None
        self.get_style_image_as = None
        self.content_image_path = None
        self.style_image_path = None
        self.content_image_array = None
        self.style_image_array = None

    def prepare_train_data(
            self,
            get_content_image_from='directory',
            content_image_path=None,
            content_image_array=None,
            get_style_image_as='filepath',
            style_image_path=None,
            style_image_array=None):
        assert get_content_image_from in ["directory", "argument"], \
            "get_content_image_from should be either 'directory' or 'argument'."
        assert get_style_image_as in ["filepath", "ndarray"], \
            "get_style_image_as should be either 'filepath' or 'ndarray'."

        self.get_content_image_from = get_content_image_from
        self.get_style_image_as = get_style_image_as
        self.content_image_path = content_image_path
        self.style_image_path = style_image_path
        self.content_image_array = content_image_array
        self.style_image_array = style_image_array

    def generate(
            self,
            iterations=10,
            output_type="file",
            output_path="result",
            total_variation_weight=1.0,
            style_weight=1.0,
            content_weight=0.025):
        self.total_variation_weight = total_variation_weight
        self.style_weight = style_weight
        self.content_weight = content_weight
        if self.get_style_image_as == 'filepath':
            try:
                load_img(self.style_image_path)
                style_image = self.style_image_path
            except Exception as e:
                raise e
        else:
            if self.style_image_array.ndim == 3:
                style_image_array = np.array([self.style_image_array])
            elif self.style_image_array.ndim < 4:
                raise ValueError("style_image_array.ndim is less than 3")
            style_image = style_image_array

        if self.get_content_image_from == "directory":
            if not os.path.isdir(self.content_image_path):
                raise FileNotFoundError(
                    self.content_image_path + " is not an existing directory.")
            parent = os.listdir(self.content_image_path)
        else:
            if self.content_image_array.ndim == 3:
                content_image_array = np.array([self.content_image_array])
            elif self.content_image_array.ndim < 4:
                raise ValueError("content_image_array.ndim is less than 3")
            parent = content_image_array

        assert output_type in ["file", "array"], \
            "output_type should be either 'file' or 'array'."

        self.result = None
        for child in parent:
            self.generate_one_image(self.content_image_path, child, style_image,
                                    iterations, output_type, output_path)

        if output_type == 'array':
            return self.result
        else:
            return

    def generate_one_image(self, content_image_path, child, style_image,
                           iterations, output_type, output_path):
        # dimensions of the generated picture.
        if self.get_content_image_from == 'directory':
            width, height = load_img(content_image_path + "/" + child).size
        elif self.get_content_image_from == 'argument':
            if K.image_data_format() == 'channels_last':
                width = child[0]
                height = child[1]
            else:
                width = child[1]
                height = child[2]
        img_nrows = 400
        img_ncols = int(width * img_nrows / height)

        x, model = self.create_model(img_nrows, img_ncols, content_image_path,
                                     child, style_image)

        loss = self.make_loss(model, img_nrows, img_ncols)

        self.set_f_outputs(loss)

        # run scipy-based optimization (L-BFGS) over the pixels of the generated image
        # so as to minimize the neural style loss
        for i in range(iterations):
            print('Start of iteration', i)
            start_time = time.time()
            x, min_val, _ = fmin_l_bfgs_b(self.loss, x.flatten(),
                                          args=(img_nrows, img_ncols),
                                          fprime=self.grads, maxfun=20)
            print('Current loss value:', min_val)
            end_time = time.time()
            print('Iteration %d completed in %ds' % (i, end_time - start_time))

        img = self.deprocess_image(x.copy(), img_nrows, img_ncols)
        if output_type == 'array':
            if self.result is None:
                img = cv2.resize(img, (224, 224))
                self.result = np.array([img])
            else:
                img = cv2.resize(img, (224, 224))
                img = np.array([img])
                self.result = np.concatenate((self.result, img))
        elif output_type == 'file':
            # save current generated image
            fname = os.getcwd() + "/" + output_path + "/" + child
            if not os.path.exists(os.path.dirname(fname)):
                os.makedirs(os.path.dirname(fname))
            skimage.io.imsave(fname, img)
            print('Image saved as', fname)

    def create_model(self, img_nrows, img_ncols, content_image_path, child,
                     style_image):
        model_type = keras.applications.vgg19

        # get tensor representations of our images
        if self.get_content_image_from == 'directory':
            x = self.preprocess_image(content_image_path + "/" + child,
                                      img_nrows, img_ncols, model_type)
        elif self.get_content_image_from == 'argument':
            x = self.preprocess_image(child, img_nrows, img_ncols, model_type)
        content_image = K.variable(x)
        style_image_tensor = K.variable(
            self.preprocess_image(style_image, img_nrows, img_ncols,
                                  model_type))

        # this will contain our generated image
        if K.image_data_format() == 'channels_first':
            self.combination_image = K.placeholder((1, 3, img_nrows, img_ncols))
        else:
            self.combination_image = K.placeholder((1, img_nrows, img_ncols, 3))

        # combine the 3 images into a single Keras tensor
        input_tensor = K.concatenate(
            [content_image, style_image_tensor, self.combination_image], axis=0)

        # build the VGG16 network with our 3 images as input
        # the model will be loaded with pre-trained ImageNet weights
        model = keras.applications.vgg19.VGG19(input_tensor=input_tensor,
                                               weights='imagenet',
                                               include_top=False)
        print('Model loaded.')

        return x, model

    def set_f_outputs(self, loss):
        # get the gradients of the generated image wrt the loss
        grads = K.gradients(loss, self.combination_image)

        outputs = [loss]
        if isinstance(grads, (list, tuple)):
            outputs += grads
        else:
            outputs.append(grads)

        self.f_outputs = K.function([self.combination_image], outputs)

    def preprocess_image(self, image_path, img_nrows, img_ncols, model_type):
        if not isinstance(image_path, np.ndarray):
            img = load_img(image_path, target_size=(img_nrows, img_ncols))
            img = img_to_array(img)
            img = np.expand_dims(img, axis=0)
        img = model_type.preprocess_input(img)
        return img

    def deprocess_image(self, x, img_nrows, img_ncols):
        if K.image_data_format() == 'channels_first':
            x = x.reshape((3, img_nrows, img_ncols))
            x = x.transpose((1, 2, 0))
        else:
            x = x.reshape((img_nrows, img_ncols, 3))
        # Remove zero-center by mean pixel
        x[:, :, 0] += 103.939
        x[:, :, 1] += 116.779
        x[:, :, 2] += 123.68
        # 'BGR'->'RGB'
        x = x[:, :, ::-1]
        x = np.clip(x, 0, 255).astype('uint8')
        return x

    # the gram matrix of an image tensor (feature-wise outer product)
    def gram_matrix(self, x):
        assert K.ndim(x) == 3
        if K.image_data_format() == 'channels_first':
            features = K.batch_flatten(x)
        else:
            features = K.batch_flatten(K.permute_dimensions(x, (2, 0, 1)))
        gram = K.dot(features, K.transpose(features))
        return gram

    # an auxiliary loss function
    # designed to maintain the "content" of the
    # content image in the generated image
    def content_loss(self, base, combination):
        return K.sum(K.square(combination - base))

    # the "style loss" is designed to maintain
    # the style of the style image in the generated image.
    # It is based on the gram matrices (which capture style) of
    # feature maps from the style style image
    # and from the generated image
    def style_loss(self, style, combination, img_nrows, img_ncols):
        assert K.ndim(style) == 3
        assert K.ndim(combination) == 3
        S = self.gram_matrix(style)
        C = self.gram_matrix(combination)
        channels = 3
        size = img_nrows * img_ncols
        return K.sum(K.square(S - C)) / (4. * (channels ** 2) * (size ** 2))

    # the 3rd loss function, total variation loss,
    # designed to keep the generated image locally coherent
    def total_variation_loss(self, x, img_nrows, img_ncols):
        assert K.ndim(x) == 4
        if K.image_data_format() == 'channels_first':
            a = K.square(
                x[:, :, :img_nrows - 1, :img_ncols - 1]
                - x[:, :, 1:, :img_ncols - 1]
            )
            b = K.square(
                x[:, :, :img_nrows - 1, :img_ncols - 1]
                - x[:, :, :img_nrows - 1, 1:]
            )
        else:
            a = K.square(
                x[:, :img_nrows - 1, :img_ncols - 1, :]
                - x[:, 1:, :img_ncols - 1, :]
            )
            b = K.square(
                x[:, :img_nrows - 1, :img_ncols - 1, :]
                - x[:, :img_nrows - 1, 1:, :]
            )
        return K.sum(K.pow(a + b, 1.25))

    def make_loss(self, model, img_nrows, img_ncols):
        """
        get the symbolic outputs of each "key" layer
        (we gave them unique names).
        """
        outputs_dict = {layer.name: layer.output for layer in model.layers}

        # combine these loss functions into a single scalar
        loss = K.variable(0.)
        layer_features = outputs_dict['block5_conv2']
        content_image_features = layer_features[0, :, :, :]
        combination_features = layer_features[2, :, :, :]
        loss += self.content_weight * self.content_loss(content_image_features,
                                                        combination_features)

        feature_layers = ['block1_conv1', 'block2_conv1', 'block3_conv1',
                          'block4_conv1', 'block5_conv1']
        for layer_name in feature_layers:
            layer_features = outputs_dict[layer_name]
            style_reference_features = layer_features[1, :, :, :]
            combination_features = layer_features[2, :, :, :]
            sl = self.style_loss(style_reference_features, combination_features,
                                 img_nrows, img_ncols)
            loss += (self.style_weight / len(feature_layers)) * sl
        loss += self.total_variation_weight * self.total_variation_loss(
            self.combination_image, img_nrows, img_ncols)
        return loss

    def loss(self, x, img_nrows, img_ncols):
        assert self.loss_value is None
        loss_value, grad_values = self.eval_loss_and_grads(x, img_nrows,
                                                           img_ncols)
        self.loss_value = loss_value
        self.grad_values = grad_values
        return self.loss_value

    def grads(self, x, img_nrows, img_ncols):
        assert self.loss_value is not None
        grad_values = np.copy(self.grad_values)
        self.loss_value = None
        self.grad_values = None
        return grad_values

    def eval_loss_and_grads(self, x, img_nrows, img_ncols):
        if K.image_data_format() == 'channels_first':
            x = x.reshape((1, 3, img_nrows, img_ncols))
        else:
            x = x.reshape((1, img_nrows, img_ncols, 3))
        outs = self.f_outputs([x])
        loss_value = outs[0]
        if len(outs[1:]) == 1:
            grad_values = outs[1].flatten().astype('float64')
        else:
            grad_values = np.array(outs[1:]).flatten().astype('float64')
        return loss_value, grad_values



import json

from keras.engine import saving
from keras.utils.io_utils import H5Dict

HDF5_OBJECT_HEADER_LIMIT = 64512


def save_models_with_class_indices(model, filepath, include_optimizer=True):
    h5dict = H5Dict(filepath, mode='w')
    h5dict['class_indices'] = json.dumps(model.class_indices)
    try:
        saving._serialize_model(model, h5dict, include_optimizer)
    finally:
        h5dict.close()


def load_models_with_class_indices(filepath, custom_objects=None, compile=True):
    model = None
    h5dict = H5Dict(filepath, 'r')
    try:
        model = saving._deserialize_model(h5dict, custom_objects, compile)
        class_indices_json = h5dict['class_indices']
        model.class_indices = json.loads(class_indices_json)
    finally:
        h5dict.close()

    return model



"""
Mask R-CNN
Multi-GPU Support for Keras.

Copyright (c) 2017 Matterport, Inc.
Licensed under the MIT License (see LICENSE for details)
Written by Waleed Abdulla

Ideas and a small code snippets from these sources:
https://github.com/fchollet/keras/issues/2436
https://medium.com/@kuza55/transparent-multi-gpu-training-on-tensorflow-with-keras-8b0016fd9012
https://github.com/avolkov1/keras_experiments/blob/master/keras_exp/multigpu/
https://github.com/fchollet/keras/blob/master/keras/utils/training_utils.py

Modification from omnis: Removed duplicate class.
"""

import keras.backend as K
import keras.layers as KL
import keras.models as KM
import tensorflow as tf


class ParallelModel(KM.Model):
    """Subclasses the standard Keras Model and adds multi-GPU support.
    It works by creating a copy of the model on each GPU. Then it slices
    the inputs and sends a slice to each copy of the model, and then
    merges the outputs together and applies the loss on the combined
    outputs.
    """
    def __init__(self, keras_model, gpu_count):
        """Class constructor.
        keras_model: The Keras model to parallelize
        gpu_count: Number of GPUs. Must be > 1
        """
        super(ParallelModel, self).__init__()
        self.inner_model = keras_model
        self.gpu_count = gpu_count
        merged_outputs = self.make_parallel()
        super(ParallelModel, self).__init__(inputs=self.inner_model.inputs,
                                            outputs=merged_outputs)

    def __getattribute__(self, attrname):
        """Redirect loading and saving methods to the inner model. That's where
        the weights are stored."""
        if 'load' in attrname or 'save' in attrname:
            return getattr(self.inner_model, attrname)
        return super(ParallelModel, self).__getattribute__(attrname)

    def summary(self, *args, **kwargs):
        """Override summary() to display summaries of both, the wrapper
        and inner models."""
        super(ParallelModel, self).summary(*args, **kwargs)
        self.inner_model.summary(*args, **kwargs)

    def make_parallel(self):
        """Creates a new wrapper model that consists of multiple replicas of
        the original model placed on different GPUs.
        """
        # Slice inputs. Slice inputs on the CPU to avoid sending a copy
        # of the full inputs to all GPUs. Saves on bandwidth and memory.
        input_slices = {name: tf.split(x, self.gpu_count)
                        for name, x in zip(self.inner_model.input_names,
                                           self.inner_model.inputs)}

        output_names = self.inner_model.output_names
        outputs_all = []
        for i in range(len(self.inner_model.outputs)):
            outputs_all.append([])

        # Run the model call() on each GPU to place the ops there
        for i in range(self.gpu_count):
            with tf.device('/gpu:%d' % i):
                with tf.name_scope('tower_%d' % i):
                    # Run a slice of inputs through this replica
                    zipped_inputs = zip(self.inner_model.input_names,
                                        self.inner_model.inputs)
                    inputs = [
                        KL.Lambda(
                            lambda s: input_slices[name][i],
                            output_shape=lambda s: (None,) + s[1:]
                        )(tensor)
                        for name, tensor in zipped_inputs]
                    # Create the model replica and get the outputs
                    outputs = self.inner_model(inputs)
                    if not isinstance(outputs, list):
                        outputs = [outputs]
                    # Save the outputs for merging back together later
                    for l, o in enumerate(outputs):
                        outputs_all[l].append(o)

        # Merge outputs on CPU
        with tf.device('/cpu:0'):
            merged = []
            for outputs, name in zip(outputs_all, output_names):
                # Concatenate or average outputs?
                # Outputs usually have a batch dimension and we concatenate
                # across it. If they don't, then the output is likely a loss
                # or a metric value that gets averaged across the batch.
                # Keras expects losses and metrics to be scalars.
                if K.int_shape(outputs[0]) == ():
                    # Average
                    m = KL.Lambda(lambda o: tf.add_n(
                        o) / len(outputs), name=name)(outputs)
                else:
                    # Concatenate
                    m = KL.Concatenate(axis=0, name=name)(outputs)
                merged.append(m)
        return merged


if __name__ == "__main__":
    # Testing code below. It creates a simple model to train on MNIST and
    # tries to run it on 2 GPUs. It saves the graph so it can be viewed
    # in TensorBoard. Run it as:
    #
    # python3 parallel_model.py

    import os
    import numpy as np
    import keras.optimizers
    from keras.datasets import mnist
    from keras.preprocessing.image import ImageDataGenerator

    GPU_COUNT = 2

    # Root directory of the project
    ROOT_DIR = os.path.abspath("../")

    # Directory to save logs and trained model
    MODEL_DIR = os.path.join(ROOT_DIR, "logs")


    def build_model(x_train, num_classes):
        # Reset default graph. Keras leaves old ops in the graph,
        # which are ignored for execution but clutter graph
        # visualization in TensorBoard.
        tf.reset_default_graph()

        inputs = KL.Input(shape=x_train.shape[1:], name="input_image")
        x = KL.Conv2D(32, (3, 3), activation='relu', padding="same",
                      name="conv1")(inputs)
        x = KL.Conv2D(64, (3, 3), activation='relu', padding="same",
                      name="conv2")(x)
        x = KL.MaxPooling2D(pool_size=(2, 2), name="pool1")(x)
        x = KL.Flatten(name="flat1")(x)
        x = KL.Dense(128, activation='relu', name="dense1")(x)
        x = KL.Dense(num_classes, activation='softmax', name="dense2")(x)

        return KM.Model(inputs, x, "digit_classifier_model")


    # Load MNIST Data
    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    x_train = np.expand_dims(x_train, -1).astype('float32') / 255
    x_test = np.expand_dims(x_test, -1).astype('float32') / 255

    print('x_train shape:', x_train.shape)
    print('x_test shape:', x_test.shape)

    # Build data generator and model
    datagen = ImageDataGenerator()
    model = build_model(x_train, 10)

    # Add multi-GPU support.
    model = ParallelModel(model, GPU_COUNT)

    optimizer = keras.optimizers.SGD(lr=0.01, momentum=0.9, clipnorm=5.0)

    model.compile(loss='sparse_categorical_crossentropy',
                  optimizer=optimizer, metrics=['accuracy'])

    model.summary()

    # Train
    model.fit_generator(
        datagen.flow(x_train, y_train, batch_size=64),
        steps_per_epoch=50, epochs=10, verbose=1,
        validation_data=(x_test, y_test),
        callbacks=[keras.callbacks.TensorBoard(log_dir=MODEL_DIR,
                                               write_graph=True)]
    )



# -*- coding: utf-8 -*-
"""
Class definition of YOLO_v3 style detection model on image and video
"""

import os
from timeit import default_timer as timer

import cv2
import numpy as np
from PIL import Image, ImageDraw
from keras import backend as K
from keras.layers import Input, Lambda
from keras.models import Model
from keras.optimizers import Adam
from keras.utils import multi_gpu_model

from ...application import Application
from ....lib.config_lib import ObjectDetectionConfig
from ....lib.general_lib import get_data_path_type
from ....lib.yolo_v3.model import (
    preprocess_true_boxes, yolo_body, yolo_eval, yolo_loss,
)
from ....lib.yolo_v3.utils import get_random_data, letterbox_image


class ObjectDetection(Application):
    def __init__(self, weights_path=None, classes_path=None, model_type='yolo'):
        Application.__init__(self)

        self.model_type = model_type

        if not isinstance(weights_path, type(None)) and not isinstance(
                classes_path, type(None)):
            self.load(weights_path, classes_path)

    def train(self,
              data_path,
              classes_path,
              epochs=50,
              batch_size=32,
              learning_rate=1e-3,
              validation_proportion=0.1):

        self.class_names = self.get_classes(classes_path)
        num_classes = len(self.class_names)

        train_lines, valid_lines = self.get_train_valid_lines(data_path,
                                                              validation_proportion)
        train_generator, valid_generator = self.make_train_valid_generator(
            train_lines,
            valid_lines,
            batch_size,
            num_classes,
            validation_proportion)

        if not hasattr(self, 'model'):
            self.model = self.create_model(ObjectDetectionConfig.INPUT_SHAPE,
                                           ObjectDetectionConfig.ANCHORS,
                                           num_classes,
                                           learning_rate=learning_rate)
        self.model.compile(optimizer=Adam(lr=learning_rate),
                           loss='mean_squared_error')

        self.model.fit_generator(train_generator,
                                 steps_per_epoch=(len(
                                     train_lines) // batch_size) + 1,
                                 validation_data=valid_generator,
                                 validation_steps=(len(
                                     valid_lines) // batch_size) + 1,
                                 epochs=epochs,
                                 callbacks=[])

    def get_train_valid_lines(self, data_path, validation_proportion):
        with open(data_path) as f:
            annotation_lines = f.readlines()

        num_val = int(len(annotation_lines) * validation_proportion)
        num_train = len(annotation_lines) - num_val

        return annotation_lines[:num_train], annotation_lines[num_train:]

    def make_train_valid_generator(self,
                                   train_lines,
                                   valid_lines,
                                   batch_size,
                                   num_classes,
                                   validation_proportion):
        train_generator = self.data_generator(train_lines,
                                              batch_size,
                                              ObjectDetectionConfig.INPUT_SHAPE,
                                              ObjectDetectionConfig.ANCHORS,
                                              num_classes)
        valid_generator = self.data_generator(valid_lines,
                                              batch_size,
                                              ObjectDetectionConfig.INPUT_SHAPE,
                                              ObjectDetectionConfig.ANCHORS,
                                              num_classes)
        return train_generator, valid_generator

    def predict(self, data_path,
                result_path=ObjectDetectionConfig.RESULT_FOLDER):
        if not os.path.isdir(result_path):
            os.makedirs(result_path)

        self.sess = K.get_session()
        get_image_from = get_data_path_type(data_path)
        if get_image_from == 'image':
            image = Image.open(data_path)
            image = self._detect_image(image)
            opencv_image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
            cv2.imwrite(os.path.join(result_path, os.path.basename(data_path)),
                        opencv_image)
        elif get_image_from == 'directory':
            filenames = os.listdir(data_path)
            for filename in filenames:
                file_full_path = os.path.join(data_path, filename)
                image = Image.open(file_full_path)
                image = self._detect_image(image)
                opencv_image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
                cv2.imwrite(os.path.join(result_path, filename), opencv_image)

    def save(self, weights_path):
        self.model.save_weights(weights_path)

    def load(self, weights_path, classes_path):
        self.class_names = self.get_classes(classes_path)
        num_classes = len(self.class_names)
        self.model = self.create_model(ObjectDetectionConfig.INPUT_SHAPE,
                                       ObjectDetectionConfig.ANCHORS,
                                       num_classes)
        self.model.load_weights(weights_path, by_name=True, skip_mismatch=True)

    def create_model_body(self,
                          anchors,
                          num_classes):
        image_input = Input(shape=(None, None, 3))
        num_anchors = len(anchors)

        model_body = yolo_body(image_input, num_anchors // 3, num_classes)

        # num = len(model_body.layers)-3
        # for i in range(num):
        #     model_body.layers[i].trainable = False

        return model_body

    def create_model(self,
                     input_shape,
                     anchors,
                     num_classes,
                     learning_rate=1e-3):
        K.clear_session()  # get a new session

        self.model_body = self.create_model_body(anchors, num_classes)

        num_anchors = len(anchors)
        h, w = input_shape
        y_true = [Input(shape=(h // {0: 32, 1: 16, 2: 8}[l],
                               w // {0: 32, 1: 16, 2: 8}[l],
                               num_anchors // 3,
                               num_classes + 5)) for l in range(3)]

        model_loss = Lambda(
            yolo_loss,
            output_shape=(1,),
            name='yolo_loss',
            arguments={
                'anchors': anchors,
                'num_classes': num_classes,
                'ignore_thresh': 0.5
            })([
                *self.model_body.output,
                *y_true
            ])

        model = Model([self.model_body.input, *y_true], model_loss)

        if self.gpu_num > 1:
            model = multi_gpu_model(model, gpus=self.gpu_num)

        return model

    def data_generator(self, annotation_lines, batch_size,
                       input_shape, anchors, num_classes):
        '''data generator for fit_generator'''
        n = len(annotation_lines)
        i = 0
        while True:
            image_data = []
            box_data = []
            for b in range(batch_size):
                if i == 0:
                    np.random.shuffle(annotation_lines)
                image, box = get_random_data(
                    annotation_lines[i], input_shape, random=True)
                image_data.append(image)
                box_data.append(box)
                i = (i + 1) % n
            image_data = np.array(image_data)  # input of original yolo: image
            # output of original yolo: boxes
            box_data = np.array(box_data)
            # some kind of output description?!
            y_true = preprocess_true_boxes(
                box_data, input_shape, anchors, num_classes)
            yield [image_data, *y_true], np.zeros(batch_size)

    def get_classes(self, classes_path):
        '''loads the classes'''
        with open(classes_path) as f:
            class_names = f.readlines()
        class_names = [c.strip() for c in class_names]
        return class_names

    def _detect_image(self, image):
        input_image_shape = K.placeholder(shape=(2,))
        num_classes = len(self.class_names)

        self.boxes, self.scores, self.classes = yolo_eval(
            self.model_body.output, ObjectDetectionConfig.ANCHORS,
            num_classes, input_image_shape,
            score_threshold=0.3, iou_threshold=0.45)

        start = timer()

        boxed_image = letterbox_image(
            image, tuple(reversed(ObjectDetectionConfig.INPUT_SHAPE)))
        image_data = np.array(boxed_image, dtype='float32')

        image_data /= 255.
        image_data = np.expand_dims(image_data, 0)  # Add batch dimension.

        out_boxes, out_scores, out_classes = self.sess.run(
            [self.boxes, self.scores, self.classes],
            feed_dict={
                self.model_body.input: image_data,
                input_image_shape: [image.size[1], image.size[0]],
                K.learning_phase(): 0
            })

        thickness = (image.size[0] + image.size[1]) // 300

        for i, c in reversed(list(enumerate(out_classes))):
            predicted_class = self.class_names[c]
            box = out_boxes[i]
            score = out_scores[i]

            label = '{} {:.2f}'.format(predicted_class, score)
            draw = ImageDraw.Draw(image)
            label_size = draw.textsize(label)

            top, left, bottom, right = box
            top = max(0, np.floor(top + 0.5).astype('int32'))
            left = max(0, np.floor(left + 0.5).astype('int32'))
            bottom = min(image.size[1], np.floor(bottom + 0.5).astype('int32'))
            right = min(image.size[0], np.floor(right + 0.5).astype('int32'))

            if top - label_size[1] >= 0:
                text_origin = np.array([left, top - label_size[1]])
            else:
                text_origin = np.array([left, top + 1])

            # My kingdom for a good redistributable image drawing library.
            for i in range(thickness):
                draw.rectangle(
                    [left + i, top + i, right - i, bottom - i],
                    outline=ObjectDetectionConfig.RECTANGLE_COLOR)
            draw.rectangle(
                [tuple(text_origin), tuple(text_origin + label_size)],
                fill=ObjectDetectionConfig.RECTANGLE_COLOR)
            draw.text(list(text_origin), label,
                      fill=ObjectDetectionConfig.TEXT_COLOR)
            del draw

        end = timer()

        return image

    def detect_video(self, video_path, output_path=""):
        vid = cv2.VideoCapture(video_path)
        print('vid: ', vid)
        print('output_path: ', output_path)
        if not vid.isOpened():
            raise IOError("Couldn't open webcam or video")
        video_FourCC = cv2.VideoWriter_fourcc(*"mp4v")
        video_fps = vid.get(cv2.CAP_PROP_FPS)
        video_size = (int(vid.get(cv2.CAP_PROP_FRAME_WIDTH)),
                      int(vid.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        is_output = output_path != ""
        if is_output:
            print("!!! TYPE:", type(output_path), type(
                video_FourCC), type(video_fps), type(video_size))
            out = cv2.VideoWriter(
                output_path, video_FourCC, video_fps, video_size)
        accum_time = 0
        curr_fps = 0
        fps = "FPS: ??"
        prev_time = timer()
        while True:
            return_value, frame = vid.read()
            if not return_value:
                break
            image = Image.fromarray(frame)
            image = self._detect_image(image)
            result = np.asarray(image)
            curr_time = timer()
            exec_time = curr_time - prev_time
            prev_time = curr_time
            accum_time = accum_time + exec_time
            curr_fps = curr_fps + 1
            if accum_time > 1:
                accum_time = accum_time - 1
                fps = "FPS: " + str(curr_fps)
                curr_fps = 0
            cv2.putText(result, text=fps, org=(3, 15),
                        fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale=0.50, color=(153, 204, 153), thickness=2)
            cv2.namedWindow("result", cv2.WINDOW_NORMAL)
            cv2.imshow("result", result)
            if is_output:
                out.write(result)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break
